"""
pyrad.prod.process_intercomp_products
=====================================

Functions for obtaining Pyrad products from datasets used in the
intercomparison process

.. autosummary::
    :toctree: generated/

    generate_intercomp_products
    generate_colocated_gates_products
    generate_time_avg_products

"""

from warnings import warn

import numpy as np

from .process_vol_products import generate_vol_products

from ..io.io_aux import get_fieldname_pyart
from ..io.io_aux import get_save_dir, make_filename

from ..io.read_data_other import read_intercomp_scores_ts

from ..io.write_data import write_colocated_gates, write_colocated_data
from ..io.write_data import write_colocated_data_time_avg
from ..io.write_data import write_intercomp_scores_ts

from ..graph.plots import plot_scatter
from ..graph.plots_timeseries import plot_intercomp_scores_ts

from ..util.radar_utils import compute_2d_stats
from ..util.stat_utils import parse_math_expression


def generate_intercomp_products(dataset, prdcfg):
    """
    Generates radar intercomparison products. Accepted product types:
        'PLOT_AND_WRITE_INTERCOMP_TS': Writes statistics of radar
            intercomparison in a file and plots the time series of the
            statistics.
            User defined parameters:
                voltype: str
                    name of the pyrad variable to use, it must be available in the dataset
                add_date_in_fname: Bool
                    If true adds the year in the csv file containing the
                    statistics. Default False
                sort_by_date: Bool
                    If true sorts the statistics by date when reading the
                    csv file containing the statistics. Default False
                step: float
                    The quantization step of the data. If none it will be
                    computed using the Py-ART config file. Default None
                rewrite: Bool
                    If true rewrites the csv file containing the statistics.
                    Default False
                npoints_min: int
                    The minimum number of points to consider the statistics
                    valid and therefore use the data point in the plotting.
                    Default 0
                transform: str
                    A transform to apply to the data before computing the statistics
                    and generating the plots. Any mathematical function with argument "x"
                    is accepted for example "sin(x) + x**2" or "10 * log10(x)"
                    Default is to use no transform
                corr_min: float
                    The minimum correlation to consider the statistics
                    valid and therefore use the data point in the plotting.
                    Default 0.
        'PLOT_SCATTER_INTERCOMP': Plots a density plot with the points of
            radar 1 versus the points of radar 2
            User defined parameters:
                voltype: str
                    name of the pyrad variable to use, it must be available in the dataset
                step: float
                    The quantization step of the data. If none it will be
                    computed using the Py-ART config file. Default None
                scatter_type: str
                    Type of scatter plot. Can be a plot for each radar volume
                    (instant) or at the end of the processing period
                    (cumulative). Default is cumulative
                range_bins: list of floats
                    Range bins to radar 1 to consider, a different plot will be generated for
                    all range bins.This can be used to differentiate the analysis in different
                    radar ranges.
                    Default is to generate only one figure, regardless of the distance to radar 1.
                transform: str
                    A transform to apply to the data before computing the statistics
                    and generating the plots. Any mathematical function with argument "x"
                    is accepted for example "sin(x) + x**2" or "10 * log10(x)"
                    Default is to use no transform
        'WRITE_INTERCOMP': Writes the instantaneously intercompared data
            (gate positions, values, etc.) in a csv file.
            User defined parameters:
                voltype: str
                    name of the pyrad variable to use, it must be available in the dataset
        'WRITE_INTERCOMP_TIME_AVG': Writes the time-averaged intercompared
            data (gate positions, values, etc.) in a csv file.
            User defined parameters:
                voltype: str
                    name of the pyrad variable to use, it must be available in the dataset

    Parameters
    ----------
    dataset : tuple
        values of colocated gates dictionary

    prdcfg : dictionary of dictionaries
        product configuration dictionary of dictionaries

    Returns
    -------
    filename : str
        the name of the file created. None otherwise

    """

    dssavedir = prdcfg["dsname"]
    if "dssavename" in prdcfg:
        dssavedir = prdcfg["dssavename"]

    if prdcfg["type"] == "WRITE_INTERCOMP":
        if dataset["final"]:
            return None

        savedir = get_save_dir(
            prdcfg["basepath"],
            prdcfg["procname"],
            dssavedir,
            prdcfg["prdname"],
            timeinfo=dataset["timeinfo"],
        )

        fname = make_filename(
            "colocated_data",
            prdcfg["dstype"],
            prdcfg["voltype"],
            ["csv"],
            timeinfo=dataset["timeinfo"],
            timeformat="%Y%m%d",
        )

        fname = savedir + fname[0]

        write_colocated_data(dataset["intercomp_dict"], fname)

        print("saved colocated data file: " + fname)

        return fname

    if prdcfg["type"] == "WRITE_INTERCOMP_TIME_AVG":
        if dataset["final"]:
            return None

        savedir = get_save_dir(
            prdcfg["basepath"],
            prdcfg["procname"],
            dssavedir,
            prdcfg["prdname"],
            timeinfo=dataset["timeinfo"],
        )

        fname = make_filename(
            "colocated_data",
            prdcfg["dstype"],
            prdcfg["voltype"],
            ["csv"],
            timeinfo=dataset["timeinfo"],
            timeformat="%Y%m%d",
        )
        fname = savedir + fname[0]
        write_colocated_data_time_avg(dataset["intercomp_dict"], fname)
        print("saved colocated time averaged data file: " + fname)

        return fname

    if prdcfg["type"] == "PLOT_SCATTER_INTERCOMP":
        scatter_type = prdcfg.get("scatter_type", "cumulative")
        if scatter_type == "cumulative" and not dataset["final"]:
            return None

        timeformat = "%Y%m%d"
        if scatter_type == "instant":
            timeformat = "%Y%m%d%H%M%S"

        field_name = get_fieldname_pyart(prdcfg["voltype"])
        savedir = get_save_dir(
            prdcfg["basepath"],
            prdcfg["procname"],
            dssavedir,
            prdcfg["prdname"],
            timeinfo=dataset["timeinfo"],
        )

        transform_str = prdcfg.get("transform", "x")
        transform = parse_math_expression(transform_str)
        range_bins = prdcfg.get("range_bins", [0, np.inf])
        step = prdcfg.get("step", None)

        fname_list = []
        for i in range(len(range_bins) - 1):  # loop on range bins
            if len(range_bins) > 2:
                rangebin_info = f"range_bin_{range_bins[i]:.0f}_{range_bins[i+1]:.0f}m"
                rangebin_info_title = (
                    f"range_bin {range_bins[i]:.0f}-{range_bins[i+1]:.0f}m"
                )
            else:
                rangebin_info = ""  # only one range bin, leave empty
                rangebin_info_title = rangebin_info

            f_list = make_filename(
                "scatter",
                prdcfg["dstype"],
                prdcfg["voltype"],
                prdcfg["imgformat"],
                prdcfginfo=rangebin_info,
                timeinfo=dataset["timeinfo"],
                timeformat=timeformat,
            )

            for j, fname in enumerate(f_list):
                f_list[j] = savedir + fname

            titl = f"colocated radar gates \n {rangebin_info_title} " + dataset[
                "timeinfo"
            ].strftime(timeformat)

            selection = np.logical_and(
                dataset["intercomp_dict"]["rad1_rng"] >= range_bins[i],
                dataset["intercomp_dict"]["rad1_rng"] < range_bins[i + 1],
            )
            if not np.sum(selection):  # skip if selection empty
                continue

            hist_2d, bin_edges1, bin_edges2, stats = compute_2d_stats(
                np.ma.asarray(dataset["intercomp_dict"]["rad1_val"][selection]),
                np.ma.asarray(dataset["intercomp_dict"]["rad2_val"][selection]),
                field_name,
                field_name,
                step1=step,
                step2=step,
                transform=transform,
            )
            if hist_2d is None:
                return None

            metadata = (
                "npoints: "
                + str(stats["npoints"])
                + "\n"
                + "mode bias: "
                + "{:.2f}".format(float(stats["modebias"]))
                + "\n"
                + "median bias: "
                + "{:.2f}".format(float(stats["medianbias"]))
                + "\n"
                + "mean bias: "
                + "{:.2f}".format(float(stats["meanbias"]))
                + "\n"
                + "intercep slope 1: "
                + "{:.2f}".format(float(stats["intercep_slope_1"]))
                + "\n"
                + "corr: "
                + "{:.2f}".format(float(stats["corr"]))
                + "\n"
                + "slope: "
                + "{:.2f}".format(float(stats["slope"]))
                + "\n"
                + "intercep: "
                + "{:.2f}".format(float(stats["intercep"]))
                + "\n"
            )

            if transform_str != "x":
                field_name = f"{transform_str} of {field_name}"
            plot_scatter(
                bin_edges1,
                bin_edges2,
                np.ma.asarray(hist_2d),
                field_name,
                field_name,
                f_list,
                prdcfg,
                titl=titl,
                metadata=metadata,
                lin_regr=[stats["slope"], stats["intercep"]],
                lin_regr_slope1=stats["intercep_slope_1"],
                rad1_name=dataset["intercomp_dict"]["rad1_name"],
                rad2_name=dataset["intercomp_dict"]["rad2_name"],
            )

            fname_list.extend(f_list)

            print("----- save to " + " ".join(f_list))

        return fname_list

    if prdcfg["type"] == "PLOT_AND_WRITE_INTERCOMP_TS":
        if not dataset["final"]:
            return None

        field_name = get_fieldname_pyart(prdcfg["voltype"])
        step = prdcfg.get("step", None)
        transform = parse_math_expression(prdcfg.get("transform", "x"))
        rad1_name = dataset["intercomp_dict"]["rad1_name"]
        rad2_name = dataset["intercomp_dict"]["rad2_name"]

        hist_2d, bin_edges1, bin_edges2, stats = compute_2d_stats(
            transform(np.ma.asarray(dataset["intercomp_dict"]["rad1_val"])),
            transform(np.ma.asarray(dataset["intercomp_dict"]["rad2_val"])),
            field_name,
            field_name,
            step1=step,
            step2=step,
        )

        # put time info in file path and name
        csvtimeinfo_file = None
        timeformat = None
        sort_by_date = False
        rewrite = False
        if "add_date_in_fname" in prdcfg:
            if prdcfg["add_date_in_fname"]:
                csvtimeinfo_file = dataset["timeinfo"]
                timeformat = "%Y"
        if "sort_by_date" in prdcfg:
            sort_by_date = prdcfg["sort_by_date"]
        if "rewrite" in prdcfg:
            rewrite = prdcfg["rewrite"]

        savedir = get_save_dir(
            prdcfg["basepath"],
            prdcfg["procname"],
            dssavedir,
            prdcfg["prdname"],
            timeinfo=None,
        )

        csvfname = make_filename(
            "ts",
            prdcfg["dstype"],
            prdcfg["voltype"],
            ["csv"],
            prdcfginfo=rad1_name + "-" + rad2_name,
            timeinfo=csvtimeinfo_file,
            timeformat=timeformat,
        )[0]

        csvfname = savedir + csvfname

        write_intercomp_scores_ts(
            dataset["timeinfo"],
            stats,
            field_name,
            csvfname,
            rad1_name=rad1_name,
            rad2_name=rad2_name,
        )
        print("saved CSV file: " + csvfname)

        (
            date_vec,
            np_vec,
            meanbias_vec,
            medianbias_vec,
            quant25bias_vec,
            quant75bias_vec,
            modebias_vec,
            corr_vec,
            slope_vec,
            intercep_vec,
            intercep_slope1_vec,
        ) = read_intercomp_scores_ts(csvfname, sort_by_date=sort_by_date)

        if date_vec is None:
            warn("Unable to plot time series. No valid data")
            return None

        if len(date_vec) < 2:
            warn("Unable to plot time series. Not enough points")
            return None

        if rewrite:
            stats = {
                "npoints": np_vec,
                "meanbias": meanbias_vec,
                "medianbias": medianbias_vec,
                "quant25bias": quant25bias_vec,
                "quant75bias": quant75bias_vec,
                "modebias": modebias_vec,
                "corr": corr_vec,
                "slope": slope_vec,
                "intercep": intercep_vec,
                "intercep_slope_1": intercep_slope1_vec,
            }
            write_intercomp_scores_ts(
                date_vec,
                stats,
                field_name,
                csvfname,
                rad1_name=rad1_name,
                rad2_name=rad2_name,
            )

        figtimeinfo = None
        titldate = (
            date_vec[0].strftime("%Y%m%d") + "-" + date_vec[-1].strftime("%Y%m%d")
        )
        if "add_date_in_fname" in prdcfg:
            if prdcfg["add_date_in_fname"]:
                figtimeinfo = date_vec[0]
                timeformat = "%Y"

        figfname_list = make_filename(
            "ts",
            prdcfg["dstype"],
            prdcfg["voltype"],
            prdcfg["imgformat"],
            prdcfginfo=rad1_name + "-" + rad2_name,
            timeinfo=figtimeinfo,
            timeformat=timeformat,
        )

        for i, figfname in enumerate(figfname_list):
            figfname_list[i] = savedir + figfname

        np_min = prdcfg.get("npoints_min", 0)
        corr_min = prdcfg.get("corr_min", 0.0)

        titl = (
            rad1_name
            + "-"
            + rad2_name
            + " "
            + field_name
            + " intercomparison "
            + titldate
        )
        plot_intercomp_scores_ts(
            date_vec,
            np_vec,
            meanbias_vec,
            medianbias_vec,
            quant25bias_vec,
            quant75bias_vec,
            modebias_vec,
            corr_vec,
            slope_vec,
            intercep_vec,
            intercep_slope1_vec,
            figfname_list,
            ref_value=0.0,
            np_min=np_min,
            corr_min=corr_min,
            labelx="Time UTC",
            titl=titl,
        )
        print("----- save to " + " ".join(figfname_list))

        return figfname_list

    warn(" Unsupported product type: " + prdcfg["type"])
    return None


def generate_colocated_gates_products(dataset, prdcfg):
    """
    Generates colocated gates products. Accepted product types:
        WRITE_COLOCATED_GATES: Writes the position of the co-located gates
            in a csv file
        All the products of the 'VOL' dataset group


    Parameters
    ----------
    dataset : tuple
        radar objects and colocated gates dictionary

    prdcfg : dictionary of dictionaries
        product configuration dictionary of dictionaries

    Returns
    -------
    filename : str
        the name of the file created. None otherwise

    """
    if prdcfg["type"] == "WRITE_COLOCATED_GATES":
        if prdcfg["radar"] not in dataset:
            return None
        if "coloc_dict" not in dataset[prdcfg["radar"]]:
            return None

        savedir = get_save_dir(
            prdcfg["basepath"],
            prdcfg["procname"],
            "colocated_gates",
            prdcfg["prdname"],
            timeinfo=None,
        )

        fname = make_filename(
            "info", prdcfg["dstype"], prdcfg["prdname"], ["csv"], timeinfo=None
        )

        fname = savedir + fname[0]

        write_colocated_gates(dataset[prdcfg["radar"]]["coloc_dict"], fname)

        print("saved colocated gates file: " + fname)

        return fname

    if prdcfg["radar"] not in dataset:
        return None
    if "radar_out" not in dataset[prdcfg["radar"]]:
        return None

    prdcfg["timeinfo"] = None
    return generate_vol_products(dataset[prdcfg["radar"]], prdcfg)


def generate_time_avg_products(dataset, prdcfg):
    """
    generates time average products. Accepted product types:
        All the products of the 'VOL' dataset group

    Parameters
    ----------
    dataset : tuple
        radar objects and colocated gates dictionary

    prdcfg : dictionary of dictionaries
        product configuration dictionary of dictionaries

    Returns
    -------
    filename : str
        the name of the file created. None otherwise

    """
    prdcfg["timeinfo"] = dataset["timeinfo"]

    return generate_vol_products(dataset, prdcfg)



"""
pyrad.graph.plots_aux
=====================

Auxiliary plotting functions

.. autosummary::
    :toctree: generated/

    generate_complex_range_Doppler_title
    generate_angle_Doppler_title
    generate_complex_Doppler_title
    generate_fixed_rng_span_title
    generate_fixed_rng_title
    generate_dda_map_title
    generate_dda_latitude_slice_title
    generate_dda_longitude_slice_title
    get_colobar_label
    get_field_name
    get_norm

"""

import numpy as np

import pyart

import matplotlib as mpl
import matplotlib.cm

mpl.use("Agg")

# Increase a bit font size
mpl.rcParams.update({"font.size": 16})
mpl.rcParams.update({"font.family": "sans-serif"})


def generate_complex_range_Doppler_title(radar, field, ray, datetime_format=None):
    """
    creates the fixed range plot title

    Parameters
    ----------
    radar : radar
        The radar object
    field : str
        name of the field
    stat : str
        The statistic computed
    datetime_forat : str or None
        The date time format to use

    Returns
    -------
    titl : str
        The plot title

    """
    begin_time = pyart.graph.common.generate_radar_time_begin(radar)
    if datetime_format:
        time_str = begin_time.strftime(datetime_format)
    else:
        time_str = begin_time.isoformat() + "Z"
    l1 = "%s azi%.1f-ele%.1f deg. %s " % (
        pyart.graph.common.generate_radar_name(radar),
        radar.azimuth["data"][ray],
        radar.elevation["data"][ray],
        time_str,
    )
    field_name = pyart.graph.common.generate_field_name(radar, field)
    return l1 + "\n" + field_name


def generate_angle_Doppler_title(
    radar, field, ang, ind_rng, along_azi=True, datetime_format=None
):
    """
    creates the angle-Doppler plot title

    Parameters
    ----------
    radar : radar
        The radar object
    field : str
        name of the field
    ang : float
        The fixed angle
    ind_rng : int
        the index of the fixed range
    along_azi : bool
        If true the plot is performed along azimuth, otherwise it is performed
        along elevation
    datetime_forat : str or None
        The date time format to use

    Returns
    -------
    titl : str
        The plot title

    """
    begin_time = pyart.graph.common.generate_radar_time_begin(radar)
    if datetime_format:
        time_str = begin_time.strftime(datetime_format)
    else:
        time_str = begin_time.isoformat() + "Z"
    if along_azi:
        ang_type = "ele"
    else:
        ang_type = "azi"
    l1 = "%s %s%.1f deg-rng%.1f m. %s " % (
        pyart.graph.common.generate_radar_name(radar),
        ang_type,
        ang,
        radar.range["data"][ind_rng],
        time_str,
    )
    field_name = pyart.graph.common.generate_field_name(radar, field)
    return l1 + "\n" + field_name


def generate_complex_Doppler_title(radar, field, ray, rng, datetime_format=None):
    """
    creates the fixed range plot title

    Parameters
    ----------
    radar : radar
        The radar object
    field : str
        name of the field
    stat : str
        The statistic computed
    datetime_forat : str or None
        The date time format to use

    Returns
    -------
    titl : str
        The plot title

    """
    begin_time = pyart.graph.common.generate_radar_time_begin(radar)
    if datetime_format:
        time_str = begin_time.strftime(datetime_format)
    else:
        time_str = begin_time.isoformat() + "Z"
    l1 = "%s azi%.1f-ele%.1f deg rng%.1f km. %s " % (
        pyart.graph.common.generate_radar_name(radar),
        radar.azimuth["data"][ray],
        radar.elevation["data"][ray],
        radar.range["data"][rng] / 1000.0,
        time_str,
    )
    field_name = pyart.graph.common.generate_field_name(radar, field)
    return l1 + "\n" + field_name


def generate_fixed_rng_span_title(radar, field, stat, datetime_format=None):
    """
    creates the fixed range plot title

    Parameters
    ----------
    radar : radar
        The radar object
    field : str
        name of the field
    stat : str
        The statistic computed
    datetime_forat : str or None
        The date time format to use

    Returns
    -------
    titl : str
        The plot title

    """
    begin_time = pyart.graph.common.generate_radar_time_begin(radar)
    if datetime_format:
        time_str = begin_time.strftime(datetime_format)
    else:
        time_str = begin_time.isoformat() + "Z"
    l1 = "%s %.1f-%.1f m %s. %s " % (
        pyart.graph.common.generate_radar_name(radar),
        np.min(radar.range["data"]),
        np.max(radar.range["data"]),
        stat,
        time_str,
    )
    field_name = pyart.graph.common.generate_field_name(radar, field)
    return l1 + "\n" + field_name


def generate_fixed_rng_title(radar, field, fixed_rng, datetime_format=None):
    """
    creates the fixed range plot title

    Parameters
    ----------
    radar : radar
        The radar object
    field : str
        name of the field
    fixed_rng : float
        The fixed range [m]
    datetime_forat : str or None
        The date time format to use

    Returns
    -------
    titl : str
        The plot title

    """
    begin_time = pyart.graph.common.generate_radar_time_begin(radar)
    if datetime_format:
        time_str = begin_time.strftime(datetime_format)
    else:
        time_str = begin_time.isoformat() + "Z"
    l1 = "%s %.1f m. %s " % (
        pyart.graph.common.generate_radar_name(radar),
        fixed_rng,
        time_str,
    )
    field_name = pyart.graph.common.generate_field_name(radar, field)
    return l1 + "\n" + field_name


def generate_dda_map_title(grid, field, level, datetime_format=None):
    """
    creates the dda map plot title

    Parameters
    ----------
    grid : grid
        The grid object
    field : str
        name of the background field
    level : int
        Verical level plotted.
    datetime_format : str or None
        The date time format to use

    Returns
    -------
    titl : str
        The plot title

    """
    begin_time = pyart.graph.common.generate_radar_time_begin(grid)
    if datetime_format:
        time_str = begin_time.strftime(datetime_format)
    else:
        time_str = begin_time.isoformat() + "Z"
    radar_names = ""

    for mdata in grid.metadata["additional_radars"]:
        if isinstance(mdata["radar_name"], bytes):
            mdata["radar_name"] = mdata["radar_name"].decode("utf-8")
        radar_names += "-" + mdata["radar_name"]
    height = grid.z["data"][level] / 1000.0
    l1 = f"DDA: {radar_names} {height:.1f} km {time_str}"
    field_name = "Hor. wind vectors (u,v) with " + field
    return l1 + "\n" + field_name


def generate_dda_latitude_slice_title(
    grid, field, level, datetime_format=None, wind_vectors="hor"
):
    """
    creates the dda latitude slice plot title

    Parameters
    ----------
    grid : grid
        The grid object
    field : str
        name of the background field
    level : int
        latitudinal level plotted.
    datetime_format : str or None
        The date time format to use
    wind_vectors : str
        'hor' if horizontal wind vectors are displayed (u and v) or 'ver'
        if vertical wind vectors are displayed (v and w)

    Returns
    -------
    titl : str
        The plot title

    """
    begin_time = pyart.graph.common.generate_radar_time_begin(grid)
    if datetime_format:
        time_str = begin_time.strftime(datetime_format)
    else:
        time_str = begin_time.isoformat() + "Z"
    radar_names = ""

    for mdata in grid.metadata["additional_radars"]:
        if isinstance(mdata["radar_name"], bytes):
            mdata["radar_name"] = mdata["radar_name"].decode("utf-8")
        radar_names += "-" + mdata["radar_name"]
    disp = grid.x["data"][level] / 1000.0
    if disp >= 0:
        direction = "east"
    else:
        direction = "west"
        disp = -disp

    l1 = f"DDA: {radar_names} {disp:.1f} km {direction} of origin {time_str}"
    if wind_vectors == "hor":
        field_name = "Hor. wind vectors (u,v)"
    elif wind_vectors == "ver":
        field_name = "Vert. wind vectors (v,w)"
    field_name += " with " + field
    return l1 + "\n" + field_name


def generate_dda_longitude_slice_title(
    grid, field, level, datetime_format=None, wind_vectors="hor"
):
    """
    creates the dda longitude slice plot title

    Parameters
    ----------
    grid : grid
        The grid object
    field : str
        name of the background field
    level : int
        longitudinal level plotted.
    datetime_format : str or None
        The date time format to use
    wind_vectors : str
        'hor' if horizontal wind vectors are displayed (u and v) or 'ver'
        if vertical wind vectors are displayed (v and w)

    Returns
    -------
    titl : str
        The plot title

    """
    begin_time = pyart.graph.common.generate_radar_time_begin(grid)
    if datetime_format:
        time_str = begin_time.strftime(datetime_format)
    else:
        time_str = begin_time.isoformat() + "Z"
    radar_names = ""

    for mdata in grid.metadata["additional_radars"]:
        if isinstance(mdata["radar_name"], bytes):
            mdata["radar_name"] = mdata["radar_name"].decode("utf-8")
        radar_names += "-" + mdata["radar_name"]
    disp = grid.x["data"][level] / 1000.0
    if disp >= 0:
        direction = "north"
    else:
        direction = "south"
        disp = -disp
    l1 = f"DDA: {radar_names} {disp:.1f} km {direction} of origin {time_str}"
    if wind_vectors == "hor":
        field_name = "Hor. wind vectors (u,v)"
    elif wind_vectors == "ver":
        field_name = "Vert. wind vectors (v,w)"
    field_name += " with " + field
    return l1 + "\n" + field_name


def get_colobar_label(field_dict, field_name):
    """
    creates the colorbar label using field metadata

    Parameters
    ----------
    field_dict : dict
        dictionary containing field metadata
    field_name : str
        name of the field

    Returns
    -------
    label : str
        colorbar label

    """
    if "standard_name" in field_dict:
        standard_name = field_dict["standard_name"]
    elif "long_name" in field_dict:
        standard_name = field_dict["long_name"]
    else:
        standard_name = field_name

    if "units" in field_dict:
        units = field_dict["units"]
    else:
        units = "?"

    return pyart.graph.common.generate_colorbar_label(standard_name, units)


def get_field_name(field_dict, field):
    """
    Return a nice field name for a particular field

    Parameters
    ----------
    field_dict : dict
        dictionary containing field metadata
    field : str
        name of the field

    Returns
    -------
    field_name : str
        the field name

    """
    if "standard_name" in field_dict:
        field_name = field_dict["standard_name"]
    elif "long_name" in field_dict:
        field_name = field_dict["long_name"]
    else:
        field_name = str(field)
    field_name = field_name.replace("_", " ")
    field_name = field_name[0].upper() + field_name[1:]

    return field_name


def get_norm(field_name, field_dict={}, isxarray=False):
    """
    Computes the normalization of the colormap, and gets the ticks and labels
    of the colorbar from the metadata of the field. Returns None if the
    required parameters are not present in the metadata. If field dict is
    not None the metadata is obtained directly from the field. Otherwise it is
    obtained from the Py-ART config file

    Parameters
    ----------
    field_name : str
        name of the field
    field_dict : dict or None
        dictionary containing the field and its metadata.
    isxarray : bool
        whether or not the norm will be used with xarray's plotting functions
        default is false, which means that matplotlib plotting functions will
        be used should be set to true when plotting Grid objects which are
        handled as xarray by pyart
    Returns
    -------
    norm : list
        the colormap index
    ticks : list
        the list of ticks in the colorbar
    labels : list
        the list of labels corresponding to each tick

    """
    norm = None
    ticks = None
    ticklabs = None

    ref_dict = pyart.config.get_metadata(field_name)
    cmap = mpl.colormaps.get_cmap(pyart.config.get_field_colormap(field_name))

    if field_dict is not None and "boundaries" in field_dict:
        if isxarray:
            ncolors = len(field_dict["boundaries"]) - 1
        else:
            ncolors = cmap.N
        norm = mpl.colors.BoundaryNorm(
            boundaries=field_dict["boundaries"], ncolors=ncolors
        )
    elif "boundaries" in ref_dict:
        if isxarray:
            ncolors = len(ref_dict["boundaries"]) - 1
        else:
            ncolors = cmap.N
        norm = mpl.colors.BoundaryNorm(
            boundaries=ref_dict["boundaries"], ncolors=ncolors
        )

    if field_dict is not None and "ticks" in field_dict:
        ticks = field_dict["ticks"]
        if "labels" in field_dict:
            ticklabs = field_dict["labels"]
    elif "ticks" in ref_dict:
        ticks = ref_dict["ticks"]
        if "labels" in ref_dict:
            ticklabs = ref_dict["labels"]
    return norm, ticks, ticklabs



"""
pyrad.graph.plots_grid
======================

Functions to plot data in a Cartesian grid format

.. autosummary::
    :toctree: generated/

    plot_surface
    plot_surface_raw
    plot_surface_contour
    plot_latitude_slice
    plot_longitude_slice
    plot_cross_section
    plot_dda_map
    plot_dda_slice
"""

from warnings import warn
from copy import deepcopy

import numpy as np

import matplotlib.pyplot as plt

import pyart

from .plots_aux import get_norm

try:
    import cartopy
    from cartopy.io.img_tiles import Stamen

    _CARTOPY_AVAILABLE = True
except ImportError:
    _CARTOPY_AVAILABLE = False

try:
    import pydda

    _PYDDA_AVAILABLE = True
except ImportError:
    _PYDDA_AVAILABLE = False

import matplotlib as mpl

mpl.use("Agg")

# Increase a bit font size
mpl.rcParams.update({"font.size": 16})
mpl.rcParams.update({"font.family": "sans-serif"})


def plot_surface(
    grid,
    field_name,
    level,
    prdcfg,
    fname_list,
    titl=None,
    alpha=None,
    ax=None,
    fig=None,
    display=None,
    save_fig=True,
    use_basemap=False,
):
    """
    plots a surface from gridded data

    Parameters
    ----------
    grid : Grid object
        object containing the gridded data to plot
    field_name : str
        name of the radar field to plot
    level : int
        level index
    prdcfg : dict
        dictionary containing the product configuration
    fname_list : list of str
        list of names of the files where to store the plot
    titl : str
        Plot title
    alpha : float or None
        Set the alpha transparency of the grid plot. Useful for
        overplotting radar over other datasets.
    ax : Axis
        Axis to plot on. if fig is None a new axis will be created
    fig : Figure
        Figure to add the colorbar to. If none a new figure will be created
    display : GridMapDisplay object
        The display used
    save_fig : bool
        if true save the figure. If false it does not close the plot and
        returns the handle to the figure

    Returns
    -------
    fname_list : list of str or
    fig, ax, display : tupple
        list of names of the saved plots or handle of the figure an axes

    """
    dpi = prdcfg["gridMapImageConfig"].get("dpi", 72)
    vmin = prdcfg.get("vmin", None)
    vmax = prdcfg.get("vmax", None)
    mask_outside = prdcfg.get("mask_outside", False)

    norm, ticks, ticklabs = get_norm(
        field_name, field_dict=grid.fields[field_name], isxarray=False
    )

    xsize = prdcfg["gridMapImageConfig"]["xsize"]
    ysize = prdcfg["gridMapImageConfig"]["ysize"]
    lonstep = prdcfg["gridMapImageConfig"].get("lonstep", 0.5)
    latstep = prdcfg["gridMapImageConfig"].get("latstep", 0.5)
    min_lon = prdcfg["gridMapImageConfig"].get("lonmin", 2.5)
    max_lon = prdcfg["gridMapImageConfig"].get("lonmax", 12.5)
    min_lat = prdcfg["gridMapImageConfig"].get("latmin", 43.5)
    max_lat = prdcfg["gridMapImageConfig"].get("latmax", 49.5)
    embellish = prdcfg["gridMapImageConfig"].get("embellish", True)
    colorbar_flag = prdcfg["gridMapImageConfig"].get("colorbar_flag", True)
    exact_limits = prdcfg["gridMapImageConfig"].get("exact_limits", True)

    if exact_limits:
        lon_lines = np.arange(min_lon, max_lon + lonstep, lonstep)
        lat_lines = np.arange(min_lat, max_lat + latstep, latstep)
    else:
        lon_lines = np.arange(np.floor(min_lon), np.ceil(max_lon) + 1, lonstep)
        lat_lines = np.arange(np.floor(min_lat), np.ceil(max_lat) + 1, latstep)

    if use_basemap or not _CARTOPY_AVAILABLE:
        resolution = prdcfg["gridMapImageConfig"].get("mapres", "l")
        if resolution not in ("c", "l", "i", "h", "f"):
            warn("Unknown map resolution: " + resolution)
            resolution = "l"

        if resolution == "c":
            area_thresh = 10000
        elif resolution == "l":
            area_thresh = 1000
        elif resolution == "i":
            area_thresh = 100
        elif resolution == "h":
            area_thresh = 10
        elif resolution == "f":
            area_thresh = 1
    else:
        resolution = prdcfg["gridMapImageConfig"].get("mapres", "110m")
        # Map from basemap to cartopy notation
        if resolution == "l":
            resolution = "110m"
        elif resolution == "i":
            resolution = "50m"
        elif resolution == "h":
            resolution = "10m"

        if resolution not in ("110m", "50m", "10m"):
            warn("Unknown map resolution: " + resolution)
            resolution = "110m"
        background_zoom = prdcfg["gridMapImageConfig"].get("background_zoom", 8)
        maps_list = prdcfg["gridMapImageConfig"].get("maps", [])

    if fig is None:
        fig = plt.figure(figsize=[xsize, ysize], dpi=dpi)

        if use_basemap or not _CARTOPY_AVAILABLE:
            display = pyart.graph.GridMapDisplayBasemap(grid)
            display.plot_basemap(
                lat_lines=lat_lines,
                lon_lines=lon_lines,
                resolution=resolution,
                area_thresh=area_thresh,
                auto_range=False,
                min_lon=min_lon,
                max_lon=max_lon,
                min_lat=min_lat,
                max_lat=max_lat,
            )
            display.plot_grid(
                field_name,
                level=level,
                norm=norm,
                ticks=ticks,
                title=titl,
                ticklabs=ticklabs,
                mask_outside=mask_outside,
                vmin=vmin,
                vmax=vmax,
                alpha=alpha,
                fig=fig,
            )
        else:
            projection = cartopy.crs.PlateCarree()

            display = pyart.graph.GridMapDisplay(grid)
            display.plot_grid(
                field_name,
                level=level,
                norm=norm,
                ticks=ticks,
                ticklabs=ticklabs,
                lat_lines=lat_lines,
                projection=projection,
                lon_lines=lon_lines,
                vmin=vmin,
                embellish=False,
                add_grid_lines=True,
                vmax=vmax,
                mask_outside=mask_outside,
                alpha=alpha,
                title=titl,
                colorbar_flag=colorbar_flag,
            )

            # fig = plt.gcf()
            # ax.set_extent([min_lon, max_lon, min_lat, max_lat])
            # display.plot_crosshairs(lon=lon, lat=lat)
    else:
        if use_basemap or not _CARTOPY_AVAILABLE:
            display.plot_grid(
                field_name,
                level=level,
                norm=norm,
                ticks=ticks,
                lat_lines=lat_lines,
                lon_lines=lon_lines,
                title=titl,
                ticklabs=ticklabs,
                colorbar_flag=False,
                embellish=False,
                vmin=vmin,
                vmax=vmax,
                mask_outside=mask_outside,
                alpha=alpha,
                ax=ax,
                fig=fig,
            )
        else:
            display.plot_grid(
                field_name,
                level=level,
                norm=norm,
                ticks=ticks,
                projection=ax.projection,
                lat_lines=lat_lines,
                lon_lines=lon_lines,
                ticklabs=ticklabs,
                colorbar_flag=False,
                embellish=False,
                vmin=vmin,
                vmax=vmax,
                mask_outside=mask_outside,
                alpha=alpha,
                title=titl,
                ax=ax,
                fig=fig,
            )

    if embellish:
        ax = plt.gca()
        if "relief" in maps_list:
            tiler = Stamen("terrain-background")
            projection = tiler.crs
            fig.delaxes(ax)
            ax = fig.add_subplot(111, projection=projection)
            warn(
                "The projection of the image is set to that of the "
                + "background map, i.e. "
                + str(projection),
                UserWarning,
            )

        for cartomap in maps_list:
            if cartomap == "relief":
                ax.add_image(tiler, background_zoom)
            if cartomap == "countries":
                # add countries
                countries = cartopy.feature.NaturalEarthFeature(
                    category="cultural",
                    name="admin_0_countries",
                    scale=resolution,
                    facecolor="none",
                )
                ax.add_feature(countries, edgecolor="black")
            elif cartomap == "provinces":
                # Create a feature for States/Admin 1 regions at
                # 1:resolution from Natural Earth
                states_provinces = cartopy.feature.NaturalEarthFeature(
                    category="cultural",
                    name="admin_1_states_provinces_lines",
                    scale=resolution,
                    facecolor="none",
                )
                ax.add_feature(states_provinces, edgecolor="gray")
            elif cartomap == "urban_areas" and resolution in ("10m", "50m"):
                urban_areas = cartopy.feature.NaturalEarthFeature(
                    category="cultural", name="urban_areas", scale=resolution
                )
                ax.add_feature(
                    urban_areas, edgecolor="brown", facecolor="brown", alpha=0.25
                )
            elif cartomap == "roads" and resolution == "10m":
                roads = cartopy.feature.NaturalEarthFeature(
                    category="cultural", name="roads", scale=resolution
                )
                ax.add_feature(roads, edgecolor="red", facecolor="none")
            elif cartomap == "railroads" and resolution == "10m":
                railroads = cartopy.feature.NaturalEarthFeature(
                    category="cultural", name="railroads", scale=resolution
                )
                ax.add_feature(
                    railroads, edgecolor="green", facecolor="none", linestyle=":"
                )
            elif cartomap == "coastlines":
                ax.coastlines(resolution=resolution)
            elif cartomap == "lakes":
                # add lakes
                lakes = cartopy.feature.NaturalEarthFeature(
                    category="physical", name="lakes", scale=resolution
                )
                ax.add_feature(lakes, edgecolor="blue", facecolor="blue", alpha=0.25)
            elif resolution == "10m" and cartomap == "lakes_europe":
                lakes_europe = cartopy.feature.NaturalEarthFeature(
                    category="physical", name="lakes_europe", scale=resolution
                )
                ax.add_feature(
                    lakes_europe, edgecolor="blue", facecolor="blue", alpha=0.25
                )
            elif cartomap == "rivers":
                # add rivers
                rivers = cartopy.feature.NaturalEarthFeature(
                    category="physical",
                    name="rivers_lake_centerlines",
                    scale=resolution,
                )
                ax.add_feature(rivers, edgecolor="blue", facecolor="none")
            elif resolution == "10m" and cartomap == "rivers_europe":
                rivers_europe = cartopy.feature.NaturalEarthFeature(
                    category="physical", name="rivers_europe", scale=resolution
                )
                ax.add_feature(rivers_europe, edgecolor="blue", facecolor="none")
            else:
                warn(
                    "cartomap "
                    + cartomap
                    + " for resolution "
                    + resolution
                    + " not available"
                )

    if save_fig:
        for fname in fname_list:
            fig.savefig(fname, dpi=dpi)
        plt.close(fig)

        return fname_list

    return (fig, ax, display)


def plot_surface_raw(
    grid,
    field_name,
    level,
    prdcfg,
    fname_list,
    titl=None,
    alpha=None,
    ax=None,
    fig=None,
    display=None,
    save_fig=True,
):
    """
    plots a surface from gridded data within reprojecting the data into a map

    Parameters
    ----------
    grid : Grid object
        object containing the gridded data to plot
    field_name : str
        name of the radar field to plot
    level : int
        level index
    prdcfg : dict
        dictionary containing the product configuration
    fname_list : list of str
        list of names of the files where to store the plot
    titl : str
        Plot title
    alpha : float or None
        Set the alpha transparency of the grid plot. Useful for
        overplotting radar over other datasets.
    ax : Axis
        Axis to plot on. if fig is None a new axis will be created
    fig : Figure
        Figure to add the colorbar to. If none a new figure will be created
    display : GridMapDisplay object
        The display used
    save_fig : bool
        if true save the figure. If false it does not close the plot and
        returns the handle to the figure

    Returns
    -------
    fname_list : list of str or
    fig, ax, display : tupple
        list of names of the saved plots or handle of the figure an axes

    """
    dpi = prdcfg["gridMapImageConfig"].get("dpi", 72)
    vmin = prdcfg.get("vmin", None)
    vmax = prdcfg.get("vmax", None)

    norm, ticks, ticklabs = get_norm(
        field_name, field_dict=grid.fields[field_name], isxarray=False
    )

    xsize = prdcfg["gridMapImageConfig"]["xsize"]
    ysize = prdcfg["gridMapImageConfig"]["ysize"]
    colorbar_flag = prdcfg["gridMapImageConfig"].get("colorbar_flag", True)

    if fig is None:
        fig = plt.figure(figsize=[xsize, ysize], dpi=dpi)
        ax = fig.add_subplot(111)

        display = pyart.graph.GridMapDisplay(grid)
        display.plot_grid_raw(
            field_name,
            level=level,
            norm=norm,
            ticks=ticks,
            ticklabs=ticklabs,
            vmin=vmin,
            vmax=vmax,
            alpha=alpha,
            title=titl,
            ax=ax,
            fig=fig,
            colorbar_flag=colorbar_flag,
        )
        # display.plot_crosshairs(lon=lon, lat=lat)
    else:
        display.plot_grid_raw(
            field_name,
            level=level,
            norm=norm,
            ticks=ticks,
            ticklabs=ticklabs,
            colorbar_flag=False,
            vmin=vmin,
            vmax=vmax,
            alpha=alpha,
            title=titl,
            ax=ax,
            fig=fig,
        )

    if save_fig:
        for fname in fname_list:
            fig.savefig(fname, dpi=dpi)
        plt.close(fig)

        return fname_list

    return (fig, ax, display)


def plot_surface_contour(
    grid,
    field_name,
    level,
    prdcfg,
    fname_list,
    contour_values=None,
    linewidths=1.5,
    colors="k",
    ax=None,
    fig=None,
    display=None,
    save_fig=True,
    use_basemap=False,
):
    """
    plots a contour plot from gridded data

    Parameters
    ----------
    grid : Grid object
        object containing the gridded data to plot
    field_name : str
        name of the radar field to plot
    level : int
        level index
    prdcfg : dict
        dictionary containing the product configuration
    fname_list : list of str
        list of names of the files where to store the plot
    contour_values : float array
        list of contours to plot
    linewidths : float
        width of the contour lines
    colors : color string or sequence of colors
        The contour colours
    ax : Axis
        Axis to plot on. if fig is None a new axis will be created
    fig : Figure
        Figure to add the colorbar to. If none a new figure will be created
    display : GridMapDisplay object
        The display used
    save_fig : bool
        if true save the figure if false it does not close the plot and
        returns the handle to the figure
    use_basemap : Bool
        If true uses basemap, otherwise uses cartopy.

    Returns
    -------
    fname_list : list of str or
    fig, ax : tupple
        list of names of the saved plots or handle of the figure an axes

    """
    # get contour intervals
    if contour_values is None:
        field_dict = pyart.config.get_metadata(field_name)
        if "boundaries" in field_dict:
            vmin = field_dict["boundaries"][0]
            vmax = field_dict["boundaries"][-1]
            num = len(field_dict["boundaries"])
        else:
            vmin, vmax = pyart.config.get_field_limits(field_name)
            num = 10

        contour_values = np.linspace(vmin, vmax, num=num)

    dpi = prdcfg["gridMapImageConfig"].get("dpi", 72)

    xsize = prdcfg["gridMapImageConfig"]["xsize"]
    ysize = prdcfg["gridMapImageConfig"]["ysize"]
    lonstep = prdcfg["gridMapImageConfig"].get("lonstep", 0.5)
    latstep = prdcfg["gridMapImageConfig"].get("latstep", 0.5)
    min_lon = prdcfg["gridMapImageConfig"].get("lonmin", 2.5)
    max_lon = prdcfg["gridMapImageConfig"].get("lonmax", 12.5)
    min_lat = prdcfg["gridMapImageConfig"].get("latmin", 43.5)
    max_lat = prdcfg["gridMapImageConfig"].get("latmax", 49.5)
    exact_limits = prdcfg["gridMapImageConfig"].get("exact_limits", 0)

    if fig is None:
        if exact_limits:
            lon_lines = np.arange(min_lon, max_lon + lonstep, lonstep)
            lat_lines = np.arange(min_lat, max_lat + latstep, latstep)
        else:
            lon_lines = np.arange(np.floor(min_lon), np.ceil(max_lon) + 1, lonstep)
            lat_lines = np.arange(np.floor(min_lat), np.ceil(max_lat) + 1, latstep)

        fig = plt.figure(figsize=[xsize, ysize], dpi=dpi)
        ax = fig.add_subplot(111)

        if use_basemap or not _CARTOPY_AVAILABLE:
            resolution = prdcfg["gridMapImageConfig"].get("mapres", "l")
            if resolution not in ("c", "l", "i", "h", "f"):
                warn("Unknown map resolution: " + resolution)
                resolution = "l"

            if resolution == "c":
                area_thresh = 10000
            elif resolution == "l":
                area_thresh = 1000
            elif resolution == "i":
                area_thresh = 100
            elif resolution == "h":
                area_thresh = 10
            elif resolution == "f":
                area_thresh = 1

            display = pyart.graph.GridMapDisplayBasemap(grid)
            display.plot_basemap(
                lat_lines=lat_lines,
                lon_lines=lon_lines,
                resolution=resolution,
                auto_range=False,
                area_thresh=area_thresh,
                min_lon=min_lon,
                max_lon=max_lon,
                min_lat=min_lat,
                max_lat=max_lat,
                ax=ax,
            )

            lons, lats = grid.get_point_longitude_latitude(edges=False)
            data = grid.fields[field_name]["data"][level, :, :]

            basemap = display.get_basemap()
            basemap.contour(
                lons,
                lats,
                data,
                contour_values,
                colors=colors,
                linewidths=linewidths,
                latlon=True,
            )
            ax.set_title(display.generate_grid_title(field_name, level))
        else:
            resolution = prdcfg["gridMapImageConfig"].get("mapres", "110m")
            # Map from basemap to cartopy notation
            if resolution == "l":
                resolution = "110m"
            elif resolution == "i":
                resolution = "50m"
            elif resolution == "h":
                resolution = "10m"
            if resolution not in ("110m", "50m", "10m"):
                warn("Unknown map resolution: " + resolution)
                resolution = "110m"
            background_zoom = prdcfg["gridMapImageConfig"].get("background_zoom", 8)
            maps_list = prdcfg["gridMapImageConfig"].get("maps", [])

            display = pyart.graph.GridMapDisplay(grid)
            fig, ax = display.plot_grid_contour(
                field_name,
                level=level,
                ax=ax,
                fig=fig,
                lat_lines=lat_lines,
                lon_lines=lon_lines,
                contour_values=contour_values,
                linewidths=linewidths,
                colors=colors,
                resolution=resolution,
                background_zoom=background_zoom,
                maps_list=maps_list,
            )
    else:
        if use_basemap or not _CARTOPY_AVAILABLE:
            lons, lats = grid.get_point_longitude_latitude(edges=False)
            data = grid.fields[field_name]["data"][level, :, :]

            basemap = display.get_basemap()
            basemap.contour(
                lons,
                lats,
                data,
                contour_values,
                colors=colors,
                linewidths=linewidths,
                latlon=True,
            )
        else:
            lons, lats = grid.get_point_longitude_latitude(edges=False)
            data = grid.fields[field_name]["data"][level, :, :]

            ax.contour(
                lons,
                lats,
                data,
                contour_values,
                colors=colors,
                linewidths=linewidths,
                transform=cartopy.crs.PlateCarree(),
            )
            ax.set_extent([min_lon, max_lon, min_lat, max_lat])

    if save_fig:
        for fname in fname_list:
            fig.savefig(fname, dpi=dpi)
        plt.close(fig)

        return fname_list

    return (fig, ax)


def plot_latitude_slice(grid, field_name, lon, lat, prdcfg, fname_list):
    """
    plots a latitude slice from gridded data

    Parameters
    ----------
    grid : Grid object
        object containing the gridded data to plot
    field_name : str
        name of the radar field to plot
    lon, lat : float
        coordinates of the slice to plot
    prdcfg : dict
        dictionary containing the product configuration
    fname_list : list of str
        list of names of the files where to store the plot

    Returns
    -------
    fname_list : list of str
        list of names of the created plots

    """
    dpi = 72
    if "dpi" in prdcfg["xsecImageConfig"]:
        dpi = prdcfg["xsecImageConfig"]["dpi"]

    norm, ticks, ticklabs = get_norm(
        field_name, field_dict=grid.fields[field_name], isxarray=True
    )

    xsize = prdcfg["xsecImageConfig"].get("xsize", 10.0)
    ysize = prdcfg["xsecImageConfig"].get("ysize", 5.0)
    xmin = prdcfg["xsecImageConfig"].get("xmin", None)
    xmax = prdcfg["xsecImageConfig"].get("xmax", None)
    ymin = prdcfg["xsecImageConfig"].get("ymin", None)
    ymax = prdcfg["xsecImageConfig"].get("ymax", None)
    vmin = prdcfg.get("vmin", None)
    vmax = prdcfg.get("vmax", None)

    fig = plt.figure(figsize=[xsize, ysize], dpi=dpi)
    ax = fig.add_subplot(111, aspect="equal")
    display = pyart.graph.GridMapDisplay(grid)
    display.plot_latitude_slice(
        field_name,
        lon=lon,
        lat=lat,
        norm=norm,
        colorbar_orient="horizontal",
        ticks=ticks,
        ticklabs=ticklabs,
        ax=ax,
        fig=fig,
        vmin=vmin,
        vmax=vmax,
    )
    ax.set_xlim([xmin, xmax])
    ax.set_ylim([ymin, ymax])

    for fname in fname_list:
        fig.savefig(fname, dpi=dpi)
    plt.close(fig)


def plot_longitude_slice(grid, field_name, lon, lat, prdcfg, fname_list):
    """
    plots a longitude slice from gridded data

    Parameters
    ----------
    grid : Grid object
        object containing the gridded data to plot
    field_name : str
        name of the radar field to plot
    lon, lat : float
        coordinates of the slice to plot
    prdcfg : dict
        dictionary containing the product configuration
    fname_list : list of str
        list of names of the files where to store the plot

    Returns
    -------
    fname_list : list of str
        list of names of the created plots

    """
    dpi = 72
    if "dpi" in prdcfg["xsecImageConfig"]:
        dpi = prdcfg["xsecImageConfig"]["dpi"]

    norm, ticks, ticklabs = get_norm(
        field_name, field_dict=grid.fields[field_name], isxarray=True
    )

    xsize = prdcfg["xsecImageConfig"].get("xsize", 10.0)
    ysize = prdcfg["xsecImageConfig"].get("ysize", 5.0)
    xmin = prdcfg["xsecImageConfig"].get("xmin", None)
    xmax = prdcfg["xsecImageConfig"].get("xmax", None)
    ymin = prdcfg["xsecImageConfig"].get("ymin", None)
    ymax = prdcfg["xsecImageConfig"].get("ymax", None)
    vmin = prdcfg.get("vmin", None)
    vmax = prdcfg.get("vmax", None)

    fig = plt.figure(figsize=[xsize, ysize], dpi=dpi)
    ax = fig.add_subplot(111, aspect="equal")
    display = pyart.graph.GridMapDisplay(grid)
    display.plot_longitude_slice(
        field_name,
        lon=lon,
        lat=lat,
        norm=norm,
        colorbar_orient="horizontal",
        ticks=ticks,
        ticklabs=ticklabs,
        ax=ax,
        fig=fig,
        vmin=vmin,
        vmax=vmax,
    )
    ax.set_xlim([xmin, xmax])
    ax.set_ylim([ymin, ymax])

    for fname in fname_list:
        fig.savefig(fname, dpi=dpi)
    plt.close(fig)


def plot_cross_section(grid, field_name, coord1, coord2, prdcfg, fname_list):
    """
    plots a croos section crossing two points in the grid

    Parameters
    ----------
    grid : Grid object
        object containing the gridded data to plot
    field_name : str
        name of the radar field to plot
    coord1 : tupple of floats
        lat, lon of the first point
    coord2 : tupple of floats
        lat, lon of the second point
    fname_list : list of str
        list of names of the files where to store the plot

    Returns
    -------
    fname_list : list of str
        list of names of the created plots

    """
    dpi = 72
    if "dpi" in prdcfg["xsecImageConfig"]:
        dpi = prdcfg["xsecImageConfig"]["dpi"]

    norm, ticks, ticklabs = get_norm(
        field_name, field_dict=grid.fields[field_name], isxarray=True
    )

    xsize = prdcfg["xsecImageConfig"].get("xsize", 10.0)
    ysize = prdcfg["xsecImageConfig"].get("ysize", 5.0)
    vmin = prdcfg.get("vmin", None)
    vmax = prdcfg.get("vmax", None)
    # xmin = prdcfg['xsecImageConfig'].get('xmin', None)
    # xmax = prdcfg['xsecImageConfig'].get('xmax', None)
    # ymin = prdcfg['xsecImageConfig'].get('ymin', None)
    # ymax = prdcfg['xsecImageConfig'].get('ymax', None)

    fig = plt.figure(figsize=[xsize, ysize], dpi=dpi)
    ax = fig.add_subplot(111, aspect="equal")
    display = pyart.graph.GridMapDisplay(grid)
    display.plot_cross_section(
        field_name,
        start=coord1,
        end=coord2,
        norm=norm,
        colorbar_orient="vertical",
        ticks=ticks,
        ticklabs=ticklabs,
        fig=fig,
        ax=ax,
        axislabels_flag=True,
        vmin=vmin,
        vmax=vmax,
    )
    # ax.set_ylim(
    #    [prdcfg['xsecImageConfig']['ymin'], prdcfg['xsecImageConfig']['ymax']])

    for fname in fname_list:
        fig.savefig(fname, dpi=dpi)
    plt.close(fig)


def plot_dda_map(
    grid,
    bg_field_name,
    level,
    prdcfg,
    fname_list,
    titl=None,
    alpha=None,
    ax=None,
    fig=None,
    display=None,
    save_fig=True,
    display_type="quiver",
):
    """
    This procedure plots a horizontal cross section of winds from wind fields
    generated by PyDDA.

    Parameters
    ----------
    grid : Grid object
        object containing the gridded data to plot
    bg_field_name : str
        name of the background radar field to plot (behind  the wind vectors)
    level : int
        level index
    prdcfg : dict
        dictionary containing the product configuration
    fname_list : list of str
        list of names of the files where to store the plot
    titl : str
        Plot title
    alpha : float or None
        Set the alpha transparency of the grid plot. Useful for
        overplotting radar over other datasets.
    ax : Axis
        Axis to plot on. if fig is None a new axis will be created
    fig : Figure
        Figure to add the colorbar to. If none a new figure will be created
    display : GridMapDisplay object
        The display used
    save_fig : bool
        if true save the figure. If false it does not close the plot and
        returns the handle to the figure
    display_type : str
        Display type for the wind vectors, can be either 'quiver', 'barbs' or
        'streamline'

    Returns
    -------
    fname_list : list of str or
    fig, ax, display : tupple
        list of names of the saved plots or handle of the figure an axes

    """

    if not _PYDDA_AVAILABLE:
        warn("PyDDA package not available. Unable to display wind fields")
        return None

    dpi = prdcfg["gridMapImageConfig"].get("dpi", 72)
    vmin = prdcfg.get("vmin", None)
    vmax = prdcfg.get("vmax", None)
    u_vel_contours = prdcfg.get("u_vel_contours", None)
    v_vel_contours = prdcfg.get("v_vel_contours", None)
    w_vel_contours = prdcfg.get("w_vel_contours", None)
    vector_spacing_km = prdcfg.get("vector_spacing_km", 10.0)
    quiver_len = prdcfg.get("quiver_len", 10.0)
    quiver_width = prdcfg.get("quiver_width", 0.01)
    streamline_width = prdcfg.get("streamline_width", None)
    streamline_arrowsize = prdcfg.get("streamline_arrowsize", None)
    display_type = prdcfg.get("display_type", "quiver")

    norm, ticks, ticklabs = get_norm(
        bg_field_name, field_dict=grid.fields[bg_field_name]
    )

    xsize = prdcfg["gridMapImageConfig"]["xsize"]
    ysize = prdcfg["gridMapImageConfig"]["ysize"]
    lonstep = prdcfg["gridMapImageConfig"].get("lonstep", 0.5)
    latstep = prdcfg["gridMapImageConfig"].get("latstep", 0.5)
    min_lon = prdcfg["gridMapImageConfig"].get("lonmin", 2.5)
    max_lon = prdcfg["gridMapImageConfig"].get("lonmax", 12.5)
    min_lat = prdcfg["gridMapImageConfig"].get("latmin", 43.5)
    max_lat = prdcfg["gridMapImageConfig"].get("latmax", 49.5)
    embellish = prdcfg["gridMapImageConfig"].get("embellish", True)
    exact_limits = prdcfg["gridMapImageConfig"].get("exact_limits", 0)
    colorbar_flag = prdcfg["gridMapImageConfig"].get("colorbar_flag", True)
    colorbar_contour_flag = prdcfg["gridMapImageConfig"].get(
        "colorbar_contour_flag", False
    )
    cmap = pyart.config.get_field_colormap(bg_field_name)

    if exact_limits:
        lon_lines = np.arange(min_lon, max_lon + lonstep, lonstep)
        lat_lines = np.arange(min_lat, max_lat + latstep, latstep)
    else:
        lon_lines = np.arange(np.floor(min_lon), np.ceil(max_lon) + 1, lonstep)
        lat_lines = np.arange(np.floor(min_lat), np.ceil(max_lat) + 1, latstep)

    fig = plt.figure(figsize=[xsize, ysize], dpi=dpi)

    resolution = prdcfg["gridMapImageConfig"].get("mapres", "110m")
    # Map from basemap to cartopy notation
    if resolution == "l":
        resolution = "110m"
    elif resolution == "i":
        resolution = "50m"
    elif resolution == "h":
        resolution = "10m"

    if resolution not in ("110m", "50m", "10m"):
        warn("Unknown map resolution: " + resolution)
        resolution = "110m"

    maps_list = prdcfg["gridMapImageConfig"].get("maps", [])

    if display_type == "quiver":
        ax = pydda.vis.plot_horiz_xsection_quiver_map(
            [pydda.io.read_from_pyart_grid(deepcopy(grid))],
            background_field=bg_field_name,
            level=level,
            show_lobes=True,
            bg_grid_no=0,
            vmin=vmin,
            vmax=vmax,
            u_vel_contours=u_vel_contours,
            v_vel_contours=v_vel_contours,
            w_vel_contours=w_vel_contours,
            quiver_spacing_x_km=vector_spacing_km,
            quiver_spacing_y_km=vector_spacing_km,
            quiverkey_len=quiver_len,
            quiver_width=quiver_width,
            colorbar_flag=colorbar_flag,
            colorbar_contour_flag=colorbar_contour_flag,
            u_field="eastward_wind_component",
            v_field="northward_wind_component",
            w_field="vertical_wind_component",
            title_flag=False,
            cmap=cmap,
        )

    elif display_type == "barbs":
        ax = pydda.vis.plot_horiz_xsection_barbs_map(
            [pydda.io.read_from_pyart_grid(deepcopy(grid))],
            background_field=bg_field_name,
            level=level,
            show_lobes=True,
            bg_grid_no=0,
            vmin=vmin,
            vmax=vmax,
            u_vel_contours=u_vel_contours,
            v_vel_contours=v_vel_contours,
            w_vel_contours=w_vel_contours,
            barb_spacing_x_km=vector_spacing_km,
            barb_spacing_y_km=vector_spacing_km,
            colorbar_flag=colorbar_flag,
            colorbar_contour_flag=colorbar_contour_flag,
            u_field="eastward_wind_component",
            v_field="northward_wind_component",
            w_field="vertical_wind_component",
            title_flag=False,
            cmap=cmap,
        )

    elif display_type == "streamline":
        ax = pydda.vis.plot_horiz_xsection_streamlines_map(
            [pydda.io.read_from_pyart_grid(deepcopy(grid))],
            background_field=bg_field_name,
            level=level,
            show_lobes=True,
            bg_grid_no=0,
            vmin=vmin,
            vmax=vmax,
            u_vel_contours=u_vel_contours,
            v_vel_contours=v_vel_contours,
            w_vel_contours=w_vel_contours,
            linewidth=streamline_width,
            arrowsize=streamline_arrowsize,
            colorbar_flag=colorbar_flag,
            colorbar_contour_flag=colorbar_contour_flag,
            u_field="eastward_wind_component",
            v_field="northward_wind_component",
            w_field="vertical_wind_component",
            title_flag=False,
            cmap=cmap,
        )

    # Edit parameters a posteriori
    # since they cannot be given to pyDDA
    ax.set_title(titl)
    ax.collections[0].set_alpha(alpha)
    if norm:
        ax.collections[0].set_norm(norm)
    if ticks:
        ax.collections[0].colorbar.set_ticks(ticks)
    if ticklabs:
        ax.collections[0].colorbar.set_ticklabels(ticklabs)

    ax.set_xticks(lon_lines)
    ax.set_yticks(lat_lines)
    ax.set_xlim([lon_lines[0], lon_lines[-1]])
    ax.set_ylim([lat_lines[0], lat_lines[-1]])

    if embellish and _CARTOPY_AVAILABLE:
        for cartomap in maps_list:
            if cartomap == "countries":
                # add countries
                countries = cartopy.feature.NaturalEarthFeature(
                    category="cultural",
                    name="admin_0_countries",
                    scale=resolution,
                    facecolor="none",
                )
                ax.add_feature(countries, edgecolor="black")
            elif cartomap == "provinces":
                # Create a feature for States/Admin 1 regions at
                # 1:resolution from Natural Earth
                states_provinces = cartopy.feature.NaturalEarthFeature(
                    category="cultural",
                    name="admin_1_states_provinces_lines",
                    scale=resolution,
                    facecolor="none",
                )
                ax.add_feature(states_provinces, edgecolor="gray")
            elif cartomap == "urban_areas" and resolution in ("10m", "50m"):
                urban_areas = cartopy.feature.NaturalEarthFeature(
                    category="cultural", name="urban_areas", scale=resolution
                )
                ax.add_feature(
                    urban_areas, edgecolor="brown", facecolor="brown", alpha=0.25
                )
            elif cartomap == "roads" and resolution == "10m":
                roads = cartopy.feature.NaturalEarthFeature(
                    category="cultural", name="roads", scale=resolution
                )
                ax.add_feature(roads, edgecolor="red", facecolor="none")
            elif cartomap == "railroads" and resolution == "10m":
                railroads = cartopy.feature.NaturalEarthFeature(
                    category="cultural", name="railroads", scale=resolution
                )
                ax.add_feature(
                    railroads, edgecolor="green", facecolor="none", linestyle=":"
                )
            elif cartomap == "coastlines":
                ax.coastlines(resolution=resolution)
            elif cartomap == "lakes":
                # add lakes
                lakes = cartopy.feature.NaturalEarthFeature(
                    category="physical", name="lakes", scale=resolution
                )
                ax.add_feature(lakes, edgecolor="blue", facecolor="blue", alpha=0.25)
            elif resolution == "10m" and cartomap == "lakes_europe":
                lakes_europe = cartopy.feature.NaturalEarthFeature(
                    category="physical", name="lakes_europe", scale=resolution
                )
                ax.add_feature(
                    lakes_europe, edgecolor="blue", facecolor="blue", alpha=0.25
                )
            elif cartomap == "rivers":
                # add rivers
                rivers = cartopy.feature.NaturalEarthFeature(
                    category="physical",
                    name="rivers_lake_centerlines",
                    scale=resolution,
                )
                ax.add_feature(rivers, edgecolor="blue", facecolor="none")
            elif resolution == "10m" and cartomap == "rivers_europe":
                rivers_europe = cartopy.feature.NaturalEarthFeature(
                    category="physical", name="rivers_europe", scale=resolution
                )
                ax.add_feature(rivers_europe, edgecolor="blue", facecolor="none")
            else:
                warn(
                    "cartomap "
                    + cartomap
                    + " for resolution "
                    + resolution
                    + " not available"
                )

    if save_fig:
        for fname in fname_list:
            fig.savefig(fname, dpi=dpi)
        plt.close(fig)

        return fname_list
    return (fig, ax, display)


def plot_dda_slice(
    grid,
    bg_field_name,
    slice_type,
    level,
    prdcfg,
    fname_list,
    titl=None,
    alpha=None,
    ax=None,
    fig=None,
    display=None,
    save_fig=True,
    display_type="quiver",
    wind_vectors="hor",
):
    """
    This procedure plots a cross section of wind vectors from wind fields
    generated by PyDDA in the X-Z plane at a specified latitude or longitude

        Parameters
    ----------
    grid : Grid object
        object containing the gridded data to plot
    bg_field_name : str
        name of the background radar field to plot (behind  the wind vectors)
    slice_type : str
        type of slice, can be either "latitude" or "longitude"
    level : int
        slicing level in latitudinal or longitudinal direction
    prdcfg : dict
        dictionary containing the product configuration
    fname_list : list of str
        list of names of the files where to store the plot
    titl : str
        Plot title
    alpha : float or None
        Set the alpha transparency of the grid plot. Useful for
        overplotting radar over other datasets.
    ax : Axis
        Axis to plot on. if fig is None a new axis will be created
    fig : Figure
        Figure to add the colorbar to. If none a new figure will be created
    display : GridMapDisplay object
        The display used
    save_fig : bool
        if true save the figure. If false it does not close the plot and
        returns the handle to the figure
    display_type : str
        Display type for the wind vectors, can be either 'quiver', 'barbs' or
        'streamline'
    wind_vectors : str
        'hor' if horizontal wind vectors are displayed (u and v) or 'ver'
        if vertical wind vectors are displayed (u and w) for latitude slice
        and (v and w) for longitude slice

    Returns
    -------
    fname_list : list of str or
    fig, ax, display : tupple
        list of names of the saved plots or handle of the figure an axes

    """
    if not _PYDDA_AVAILABLE:
        warn("PyDDA package not available. Unable to display wind fields")
        return None
    dpi = prdcfg["xsecImageConfig"].get("dpi", 72)
    vmin = prdcfg.get("vmin", None)
    vmax = prdcfg.get("vmax", None)
    u_vel_contours = prdcfg.get("u_vel_contours", None)
    v_vel_contours = prdcfg.get("v_vel_contours", None)
    w_vel_contours = prdcfg.get("w_vel_contours", None)
    vector_spacing_km = prdcfg.get("vector_spacing_km", 10.0)
    quiver_len = prdcfg.get("quiver_len", 10.0)
    quiver_width = prdcfg.get("quiver_width", 0.01)
    streamline_width = prdcfg.get("streamline_width", None)
    streamline_arrowsize = prdcfg.get("streamline_arrowsize", None)
    display_type = prdcfg.get("display_type", "quiver")

    norm, ticks, ticklabs = get_norm(
        bg_field_name, field_dict=grid.fields[bg_field_name]
    )

    colorbar_flag = prdcfg["xsecImageConfig"].get("colorbar_flag", True)
    colorbar_contour_flag = prdcfg["xsecImageConfig"].get(
        "colorbar_contour_flag", False
    )
    cmap = pyart.config.get_field_colormap(bg_field_name)

    xsize = prdcfg["xsecImageConfig"].get("xsize", 10.0)
    ysize = prdcfg["xsecImageConfig"].get("ysize", 5.0)
    xmin = prdcfg["xsecImageConfig"].get("xmin", None)
    xmax = prdcfg["xsecImageConfig"].get("xmax", None)
    ymin = prdcfg["xsecImageConfig"].get("ymin", None)
    ymax = prdcfg["xsecImageConfig"].get("ymax", None)

    # This is a bit hackish
    # pyDDA does not support plotting vertical profiles of u and v
    # only (v and w) or (u and w), so we trick it by assigning the v variable
    # to w

    if wind_vectors == "hor":
        w_field = "eastward_wind_component"
    else:
        w_field = "vertical_wind_component"

    fig = plt.figure(figsize=[xsize, ysize], dpi=dpi)

    if slice_type == "latitude":
        if display_type == "quiver":
            ax = pydda.vis.plot_yz_xsection_quiver(
                [pydda.io.read_from_pyart_grid(deepcopy(grid))],
                background_field=bg_field_name,
                level=level,
                bg_grid_no=0,
                vmin=vmin,
                vmax=vmax,
                u_vel_contours=u_vel_contours,
                v_vel_contours=v_vel_contours,
                w_vel_contours=w_vel_contours,
                quiver_spacing_y_km=vector_spacing_km,
                quiver_spacing_z_km=vector_spacing_km,
                quiver_width=quiver_width,
                quiverkey_len=quiver_len,
                colorbar_flag=colorbar_flag,
                colorbar_contour_flag=colorbar_contour_flag,
                u_field="eastward_wind_component",
                v_field="northward_wind_component",
                w_field=w_field,
                title_flag=False,
                cmap=cmap,
            )

        elif display_type == "barbs":
            ax = pydda.vis.plot_yz_xsection_barbs(
                [pydda.io.read_from_pyart_grid(deepcopy(grid))],
                background_field=bg_field_name,
                level=level,
                bg_grid_no=0,
                vmin=vmin,
                vmax=vmax,
                u_vel_contours=u_vel_contours,
                v_vel_contours=v_vel_contours,
                w_vel_contours=w_vel_contours,
                barb_spacing_y_km=vector_spacing_km,
                barb_spacing_z_km=vector_spacing_km,
                colorbar_flag=colorbar_flag,
                colorbar_contour_flag=colorbar_contour_flag,
                u_field="eastward_wind_component",
                v_field="northward_wind_component",
                w_field=w_field,
                title_flag=False,
                cmap=cmap,
            )

        elif display_type == "streamline":
            ax = pydda.vis.plot_yz_xsection_streamlines(
                [pydda.io.read_from_pyart_grid(deepcopy(grid))],
                background_field=bg_field_name,
                level=level,
                bg_grid_no=0,
                vmin=vmin,
                vmax=vmax,
                u_vel_contours=u_vel_contours,
                v_vel_contours=v_vel_contours,
                w_vel_contours=w_vel_contours,
                linewidth=streamline_width,
                arrowsize=streamline_arrowsize,
                colorbar_flag=colorbar_flag,
                colorbar_contour_flag=colorbar_contour_flag,
                u_field="eastward_wind_component",
                v_field="northward_wind_component",
                w_field=w_field,
                title_flag=False,
                cmap=cmap,
            )

    elif slice_type == "longitude":
        if display_type == "quiver":
            ax = pydda.vis.plot_xz_xsection_quiver(
                [pydda.io.read_from_pyart_grid(deepcopy(grid))],
                background_field=bg_field_name,
                level=level,
                bg_grid_no=0,
                vmin=vmin,
                vmax=vmax,
                u_vel_contours=u_vel_contours,
                v_vel_contours=v_vel_contours,
                w_vel_contours=w_vel_contours,
                quiver_width=quiver_width,
                quiver_spacing_x_km=vector_spacing_km,
                quiver_spacing_z_km=vector_spacing_km,
                quiverkey_len=quiver_len,
                colorbar_flag=colorbar_flag,
                colorbar_contour_flag=colorbar_contour_flag,
                u_field="eastward_wind_component",
                v_field="northward_wind_component",
                w_field=w_field,
                title_flag=False,
                cmap=cmap,
            )

        elif display_type == "barbs":
            ax = pydda.vis.plot_xz_xsection_barbs(
                [pydda.io.read_from_pyart_grid(deepcopy(grid))],
                background_field=bg_field_name,
                level=level,
                bg_grid_no=0,
                vmin=vmin,
                vmax=vmax,
                u_vel_contours=u_vel_contours,
                v_vel_contours=v_vel_contours,
                w_vel_contours=w_vel_contours,
                barb_spacing_x_km=vector_spacing_km,
                barb_spacing_z_km=vector_spacing_km,
                colorbar_flag=colorbar_flag,
                colorbar_contour_flag=colorbar_contour_flag,
                u_field="eastward_wind_component",
                v_field="northward_wind_component",
                w_field=w_field,
                title_flag=False,
                cmap=cmap,
            )

        elif display_type == "streamline":
            ax = pydda.vis.plot_xz_xsection_streamlines(
                [pydda.io.read_from_pyart_grid(deepcopy(grid))],
                background_field=bg_field_name,
                level=level,
                bg_grid_no=0,
                vmin=vmin,
                vmax=vmax,
                u_vel_contours=u_vel_contours,
                v_vel_contours=v_vel_contours,
                w_vel_contours=w_vel_contours,
                linewidth=streamline_width,
                arrowsize=streamline_arrowsize,
                colorbar_flag=colorbar_flag,
                colorbar_contour_flag=colorbar_contour_flag,
                u_field="eastward_wind_component",
                v_field="northward_wind_component",
                w_field=w_field,
                title_flag=False,
                cmap=cmap,
            )

    # Edit parameters a posteriori
    # since they cannot be given to pyDDA
    ax.set_title(titl)
    ax.collections[0].set_alpha(alpha)
    if norm:
        ax.collections[0].set_norm(norm)
    if ticks:
        ax.collections[0].colorbar.set_ticks(ticks)
    if ticklabs:
        ax.collections[0].colorbar.set_ticklabels(ticklabs)

    ax.set_xlim([xmin, xmax])
    ax.set_ylim([ymin, ymax])

    if save_fig:
        for fname in fname_list:
            fig.savefig(fname, dpi=dpi)
        plt.close(fig)

        return fname_list

    return (fig, ax, display)



"""
pyrad.proc.process_iq
=====================

Functions to processes IQ data.

.. autosummary::
    :toctree: generated/

    process_raw_iq
    process_pol_variables_iq
    process_reflectivity_iq
    process_st1_iq
    process_st2_iq
    process_wbn_iq
    process_differential_reflectivity_iq
    process_mean_phase_iq
    process_differential_phase_iq
    process_rhohv_iq
    process_Doppler_velocity_iq
    process_Doppler_width_iq
    process_fft

"""

from copy import deepcopy
from warnings import warn

import pyart

from ..io.io_aux import get_datatype_fields, get_fieldname_pyart


def process_raw_iq(procstatus, dscfg, radar_list=None):
    """
    Dummy function that returns the initial input data set

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration
    radar_list : list of spectra objects
        Optional. list of spectra objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output
    ind_rad : int
        radar index

    """

    if procstatus != 1:
        return None, None

    for datatypedescr in dscfg["datatype"]:
        radarnr, _, _, _, _ = get_datatype_fields(datatypedescr)
        break
    ind_rad = int(radarnr[5:8]) - 1
    if (radar_list is None) or (radar_list[ind_rad] is None):
        warn("ERROR: No valid radar")
        return None, None
    new_dataset = {"radar_out": deepcopy(radar_list[ind_rad])}

    return new_dataset, ind_rad


def process_pol_variables_iq(procstatus, dscfg, radar_list=None):
    """
    Computes the polarimetric variables from the IQ data

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted configuration keywords::

        datatype : list of string. Dataset keyword
            The input data types, must contain
            "IQhhADU", "IQvvADU", "IQNADUh" and "IQNADUv"
        subtract_noise : Bool
            If True noise will be subtracted from the signal
        lag : int
            The time lag to use in the estimators
        direction : str
            The convention used in the Doppler mean field. Can be
            negative_away or negative_towards
        variables : list of str
            list of variables to compute. Default dBZ
        phase_offset : float. Dataset keyword
            The system differential phase offset to remove
    radar_list : list of spectra objects
        Optional. list of spectra objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output fields corresponding to the specified
        "variables"
        ""
    ind_rad : int
        radar index

    """
    if procstatus != 1:
        return None, None

    noise_h_field = None
    noise_v_field = None
    signal_h_field = None
    signal_v_field = None
    for datatypedescr in dscfg["datatype"]:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        if datatype == "IQhhADU":
            signal_h_field = get_fieldname_pyart(datatype)
        elif datatype == "IQvvADU":
            signal_v_field = get_fieldname_pyart(datatype)
        elif datatype == "IQNADUh":
            noise_h_field = get_fieldname_pyart(datatype)
        elif datatype == "IQNADUv":
            noise_v_field = get_fieldname_pyart(datatype)

    ind_rad = int(radarnr[5:8]) - 1
    if (radar_list is None) or (radar_list[ind_rad] is None):
        warn("ERROR: No valid radar")
        return None, None
    radar = radar_list[ind_rad]

    subtract_noise = dscfg.get("subtract_noise", False)
    lag = dscfg.get("lag", None)
    direction = dscfg.get("direction", "negative_away")
    variables = dscfg.get("variables", ["dBZ"])
    phase_offset = dscfg.get("phase_offset", 0.0)

    fields_list = []
    for variable in variables:
        fields_list.append(get_fieldname_pyart(variable))

    radar = pyart.retrieve.compute_pol_variables_iq(
        radar,
        fields_list,
        subtract_noise=subtract_noise,
        lag=lag,
        direction=direction,
        phase_offset=phase_offset,
        signal_h_field=signal_h_field,
        signal_v_field=signal_v_field,
        noise_h_field=noise_h_field,
        noise_v_field=noise_v_field,
    )

    # prepare for exit
    new_dataset = {"radar_out": radar}

    return new_dataset, ind_rad


def process_reflectivity_iq(procstatus, dscfg, radar_list=None):
    """
    Computes reflectivity from the IQ data

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted configuration keywords::

        datatype : list of string. Dataset keyword
            The input data types, must contain
            "IQhhADU" or "IQvvADU", and,
            "IQNADUh" or "IQNADUv"
        subtract_noise : Bool
            If True noise will be subtracted from the signal
    radar_list : list of spectra objects
        Optional. list of spectra objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output field
        "dBZ" if "IQhhADU" and "IQNADUh" are specified
        "dBZv" if "IQvvADU" and "IQNADUv" are specified
    ind_rad : int
        radar index

    """

    if procstatus != 1:
        return None, None

    noise_field = None
    signal_field = None
    for datatypedescr in dscfg["datatype"]:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        if datatype in ("IQhhADU", "IQvvADU"):
            signal_field = get_fieldname_pyart(datatype)
        elif datatype in ("IQNADUh", "IQNADUv"):
            noise_field = get_fieldname_pyart(datatype)

    if signal_field is None:
        warn("Signal field must be specified")
        return None, None

    ind_rad = int(radarnr[5:8]) - 1
    if (radar_list is None) or (radar_list[ind_rad] is None):
        warn("ERROR: No valid radar")
        return None, None
    radar = radar_list[ind_rad]

    if signal_field not in radar.fields:
        warn("Unable to obtain reflectivity. Missing field " + signal_field)
        return None, None

    subtract_noise = dscfg.get("subtract_noise", False)

    dBZ = pyart.retrieve.compute_reflectivity_iq(
        radar,
        subtract_noise=subtract_noise,
        signal_field=signal_field,
        noise_field=noise_field,
    )

    reflectivity_field = "reflectivity"
    if signal_field in ("IQ_vv_ADU",):
        reflectivity_field += "_vv"

    # prepare for exit
    new_dataset = {"radar_out": pyart.util.radar_from_spectra(radar)}
    new_dataset["radar_out"].add_field(reflectivity_field, dBZ)

    return new_dataset, ind_rad


def process_st1_iq(procstatus, dscfg, radar_list=None):
    """
    Computes the statistical test one lag fluctuation from the horizontal or
    vertical IQ data

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted configuration keywords::

        datatype : list of string. Dataset keyword
            The input data types, must contain
            "IQhhADU" or "IQvvADU"
    radar_list : list of spectra objects
        Optional. list of spectra objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output field "ST1" (stat_test_lag1)
    ind_rad : int
        radar index

    """

    if procstatus != 1:
        return None, None

    radarnr, _, datatype, _, _ = get_datatype_fields(dscfg["datatype"][0])
    signal_field = get_fieldname_pyart(datatype)

    ind_rad = int(radarnr[5:8]) - 1
    if (radar_list is None) or (radar_list[ind_rad] is None):
        warn("ERROR: No valid radar")
        return None, None
    radar = radar_list[ind_rad]

    if signal_field not in radar.fields:
        warn("Unable to obtain ST1. Missing fields")
        return None, None

    st1 = pyart.retrieve.compute_st1_iq(radar, signal_field=signal_field)

    st1_field = "stat_test_lag1"
    if signal_field == "IQ_vv_ADU":
        st1_field += "_vv"

    # prepare for exit
    new_dataset = {"radar_out": pyart.util.radar_from_spectra(radar)}
    new_dataset["radar_out"].add_field(st1_field, st1)

    return new_dataset, ind_rad


def process_st2_iq(procstatus, dscfg, radar_list=None):
    """
    Computes the statistical test two lag fluctuation from the horizontal or
    vertical IQ data

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted configuration keywords::

        datatype : list of string. Dataset keyword
            The input data types, must contain
            "IQhhADU" or "IQvvADU"
    radar_list : list of spectra objects
        Optional. list of spectra objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output field "ST2"
    ind_rad : int
        radar index

    """

    if procstatus != 1:
        return None, None

    radarnr, _, datatype, _, _ = get_datatype_fields(dscfg["datatype"][0])
    signal_field = get_fieldname_pyart(datatype)

    ind_rad = int(radarnr[5:8]) - 1
    if (radar_list is None) or (radar_list[ind_rad] is None):
        warn("ERROR: No valid radar")
        return None, None
    radar = radar_list[ind_rad]

    if signal_field not in radar.fields:
        warn("Unable to obtain ST2. Missing fields")
        return None, None

    st2 = pyart.retrieve.compute_st2_iq(radar, signal_field=signal_field)

    st2_field = "stat_test_lag2"
    if signal_field == "IQ_vv_ADU":
        st2_field += "_vv"

    # prepare for exit
    new_dataset = {"radar_out": pyart.util.radar_from_spectra(radar)}
    new_dataset["radar_out"].add_field(st2_field, st2)

    return new_dataset, ind_rad


def process_wbn_iq(procstatus, dscfg, radar_list=None):
    """
    Computes the wide band noise from the horizontal or vertical IQ data

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted configuration keywords::

        datatype : list of string. Dataset keyword
            The input data types, must contain
            "IQhhADU" or "IQvvADU"
    radar_list : list of spectra objects
        Optional. list of spectra objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output field "WBN" (wide-band noise)
    ind_rad : int
        radar index

    """

    if procstatus != 1:
        return None, None

    radarnr, _, datatype, _, _ = get_datatype_fields(dscfg["datatype"][0])
    signal_field = get_fieldname_pyart(datatype)

    ind_rad = int(radarnr[5:8]) - 1
    if (radar_list is None) or (radar_list[ind_rad] is None):
        warn("ERROR: No valid radar")
        return None, None
    radar = radar_list[ind_rad]

    if signal_field not in radar.fields:
        warn("Unable to obtain WBN. Missing fields")
        return None, None

    wbn = pyart.retrieve.compute_wbn_iq(radar, signal_field=signal_field)

    wbn_field = "wide_band_noise"
    if signal_field == "IQ_vv_ADU":
        wbn_field += "_vv"

    # prepare for exit
    new_dataset = {"radar_out": pyart.util.radar_from_spectra(radar)}
    new_dataset["radar_out"].add_field(wbn_field, wbn)

    return new_dataset, ind_rad


def process_differential_reflectivity_iq(procstatus, dscfg, radar_list=None):
    """
    Computes differential reflectivity from the horizontal and vertical
    IQ data

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted configuration keywords::

        datatype : list of string. Dataset keyword
            The input data types, must contain
            "IQhhADU", "IQvvADU", "IQNADUh" and "IQNADUv"
        subtract_noise : Bool
            If True noise will be subtracted from the signal
        lag : int
            The time lag to use in the estimators
    radar_list : list of spectra objects
        Optional. list of spectra objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output field "ZDR"
    ind_rad : int
        radar index

    """

    if procstatus != 1:
        return None, None

    noise_h_field = None
    noise_v_field = None
    signal_h_field = None
    signal_v_field = None
    for datatypedescr in dscfg["datatype"]:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        if datatype == "IQhhADU":
            signal_h_field = get_fieldname_pyart(datatype)
        elif datatype == "IQvvADU":
            signal_v_field = get_fieldname_pyart(datatype)
        elif datatype == "IQNADUh":
            noise_h_field = get_fieldname_pyart(datatype)
        elif datatype == "IQNADUv":
            noise_v_field = get_fieldname_pyart(datatype)

    ind_rad = int(radarnr[5:8]) - 1
    if (radar_list is None) or (radar_list[ind_rad] is None):
        warn("ERROR: No valid radar")
        return None, None
    radar = radar_list[ind_rad]

    if signal_h_field not in radar.fields or signal_v_field not in radar.fields:
        warn("Unable to obtain spectral differential reflectivity. " + "Missing fields")
        return None, None

    subtract_noise = dscfg.get("subtract_noise", False)
    lag = dscfg.get("lag", 0)

    zdr = pyart.retrieve.compute_differential_reflectivity_iq(
        radar,
        subtract_noise=subtract_noise,
        lag=lag,
        signal_h_field=signal_h_field,
        signal_v_field=signal_v_field,
        noise_h_field=noise_h_field,
        noise_v_field=noise_v_field,
    )

    # prepare for exit
    new_dataset = {"radar_out": pyart.util.radar_from_spectra(radar)}
    new_dataset["radar_out"].add_field("differential_reflectivity", zdr)

    return new_dataset, ind_rad


def process_mean_phase_iq(procstatus, dscfg, radar_list=None):
    """
    Computes the mean phase from the horizontal or vertical IQ data

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted configuration keywords::

        datatype : list of string. Dataset keyword
            The input data types, must contain
            "IQhhADU" or "IQvvADU"
    radar_list : list of spectra objects
        Optional. list of spectra objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output field "MPH" (mean_phase)
    ind_rad : int
        radar index

    """

    if procstatus != 1:
        return None, None

    radarnr, _, datatype, _, _ = get_datatype_fields(dscfg["datatype"][0])
    signal_field = get_fieldname_pyart(datatype)

    ind_rad = int(radarnr[5:8]) - 1
    if (radar_list is None) or (radar_list[ind_rad] is None):
        warn("ERROR: No valid radar")
        return None, None
    radar = radar_list[ind_rad]

    if signal_field not in radar.fields:
        warn("Unable to obtain MPH. Missing fields")
        return None, None

    mph = pyart.retrieve.compute_mean_phase_iq(radar, signal_field=signal_field)

    mean_phase_field = "mean_phase"
    if signal_field == "IQ_vv_ADU":
        mean_phase_field += "_vv"

    # prepare for exit
    new_dataset = {"radar_out": pyart.util.radar_from_spectra(radar)}
    new_dataset["radar_out"].add_field(mean_phase_field, mph)

    return new_dataset, ind_rad


def process_differential_phase_iq(procstatus, dscfg, radar_list=None):
    """
    Computes the differential phase from the horizontal and vertical IQ data

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted configuration keywords::

        datatype : list of string. Dataset keyword
            The input data types, must contain
            "IQhhADU" and "IQvvADU"
        phase_offset : float. Dataset keyword
            The system differential phase offset to remove
    radar_list : list of spectra objects
        Optional. list of spectra objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output field uPhiDP
    ind_rad : int
        radar index

    """

    if procstatus != 1:
        return None, None

    signal_h_field = None
    signal_v_field = None
    for datatypedescr in dscfg["datatype"]:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        if datatype == "IQhhADU":
            signal_h_field = get_fieldname_pyart(datatype)
        elif datatype == "IQvvADU":
            signal_v_field = get_fieldname_pyart(datatype)

    ind_rad = int(radarnr[5:8]) - 1
    if (radar_list is None) or (radar_list[ind_rad] is None):
        warn("ERROR: No valid radar")
        return None, None
    radar = radar_list[ind_rad]

    if signal_h_field not in radar.fields or signal_v_field not in radar.fields:
        warn("Unable to obtain PhiDP. Missing fields")
        return None, None

    phase_offset = dscfg.get("phase_offset", 0.0)

    uphidp = pyart.retrieve.compute_differential_phase_iq(
        radar,
        phase_offset=phase_offset,
        signal_h_field=signal_h_field,
        signal_v_field=signal_v_field,
    )

    # prepare for exit
    new_dataset = {"radar_out": pyart.util.radar_from_spectra(radar)}
    new_dataset["radar_out"].add_field("uncorrected_differential_phase", uphidp)

    return new_dataset, ind_rad


def process_rhohv_iq(procstatus, dscfg, radar_list=None):
    """
    Computes RhoHV from the horizontal and vertical IQ data

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted configuration keywords::

        datatype : list of string. Dataset keyword
            The input data types, must contain
            "IQhhADU", "IQvvADU", "IQNADUh" and "IQNADUv"
        subtract_noise : Bool
            If True noise will be subtracted from the signal
        lag : int
            Time lag used in the computation
    radar_list : list of spectra objects
        Optional. list of spectra objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output field "RhoHV"
    ind_rad : int
        radar index

    """
    if procstatus != 1:
        return None, None

    noise_h_field = None
    noise_v_field = None
    signal_h_field = None
    signal_v_field = None
    for datatypedescr in dscfg["datatype"]:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        if datatype == "IQhhADU":
            signal_h_field = get_fieldname_pyart(datatype)
        elif datatype == "IQvvADU":
            signal_v_field = get_fieldname_pyart(datatype)
        elif datatype == "IQNADUh":
            noise_h_field = get_fieldname_pyart(datatype)
        elif datatype == "IQNADUv":
            noise_v_field = get_fieldname_pyart(datatype)

    ind_rad = int(radarnr[5:8]) - 1
    if (radar_list is None) or (radar_list[ind_rad] is None):
        warn("ERROR: No valid radar")
        return None, None
    radar = radar_list[ind_rad]

    if signal_h_field not in radar.fields or signal_v_field not in radar.fields:
        warn("Unable to obtain RhoHV. Missing fields")
        return None, None

    subtract_noise = dscfg.get("subtract_noise", False)
    lag = dscfg.get("lag", 0)

    rhohv = pyart.retrieve.compute_rhohv_iq(
        radar,
        subtract_noise=subtract_noise,
        lag=lag,
        signal_h_field=signal_h_field,
        signal_v_field=signal_v_field,
        noise_h_field=noise_h_field,
        noise_v_field=noise_v_field,
    )

    # prepare for exit
    new_dataset = {"radar_out": pyart.util.radar_from_spectra(radar)}
    new_dataset["radar_out"].add_field("cross_correlation_ratio", rhohv)

    return new_dataset, ind_rad


def process_Doppler_velocity_iq(procstatus, dscfg, radar_list=None):
    """
    Compute the Doppler velocity from the spectral reflectivity

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted configuration keywords::

        datatype : list of string. Dataset keyword
            The input data types, must contain
            "IQhhADU" or "IQvvADU"
        direction : str
            The convention used in the Doppler mean field. Can be
            negative_away or negative_towards
    radar_list : list of spectra objects
        Optional. list of spectra objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output field "V"
        (if IQhhADU was provided) or "Vv" (if IQvvADU was provided)
    ind_rad : int
        radar index

    """

    if procstatus != 1:
        return None, None

    for datatypedescr in dscfg["datatype"]:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        if datatype in ("IQhhADU", "IQvvADU"):
            signal_field = get_fieldname_pyart(datatype)

    ind_rad = int(radarnr[5:8]) - 1
    if (radar_list is None) or (radar_list[ind_rad] is None):
        warn("ERROR: No valid radar")
        return None, None
    radar = radar_list[ind_rad]

    if signal_field not in radar.fields:
        warn("Unable to obtain Doppler velocity. " + "Missing field " + signal_field)
        return None, None

    direction = dscfg.get("direction", "negative_away")

    vel = pyart.retrieve.compute_Doppler_velocity_iq(
        radar, signal_field=signal_field, direction=direction
    )

    vel_field = "velocity"
    if signal_field == "IQ_vv_ADU":
        vel_field += "_vv"

    # prepare for exit
    new_dataset = {"radar_out": pyart.util.radar_from_spectra(radar)}
    new_dataset["radar_out"].add_field(vel_field, vel)

    return new_dataset, ind_rad


def process_Doppler_width_iq(procstatus, dscfg, radar_list=None):
    """
    Compute the Doppler spectrum width from the spectral reflectivity

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted configuration keywords::

        datatype : list of string. Dataset keyword
            The input data types, must contain
            "IQhhADU" or "IQvvADU"
        subtract_noise : Bool
            If True noise will be subtracted from the signals
        lag : int
            Time lag used in the denominator of the computation
    radar_list : list of spectra objects
        Optional. list of spectra objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output field "W" (if IQhhADU was provided),
        or "Wv" (if IQvvADU was provided)
    ind_rad : int
        radar index

    """

    if procstatus != 1:
        return None, None

    noise_field = None
    for datatypedescr in dscfg["datatype"]:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        if datatype in ("IQhhADU", "IQvvADU"):
            signal_field = get_fieldname_pyart(datatype)
        elif datatype in ("IQNADUh", "IQNADUv"):
            noise_field = get_fieldname_pyart(datatype)

    ind_rad = int(radarnr[5:8]) - 1
    if (radar_list is None) or (radar_list[ind_rad] is None):
        warn("ERROR: No valid radar")
        return None, None
    radar = radar_list[ind_rad]

    if signal_field not in radar.fields:
        warn(
            "Unable to obtain Doppler spectrum width. "
            + "Missing field "
            + signal_field
        )
        return None, None

    dscfg.get("subtract_noise", False)
    lag = dscfg.get("lag", 1)

    width = pyart.retrieve.compute_Doppler_width_iq(
        radar,
        subtract_noise=True,
        signal_field=signal_field,
        noise_field=noise_field,
        lag=lag,
    )

    width_field = "spectrum_width"
    if signal_field == "IQ_vv_ADU":
        width_field += "_vv"

    # prepare for exit
    new_dataset = {"radar_out": pyart.util.radar_from_spectra(radar)}
    new_dataset["radar_out"].add_field(width_field, width)

    return new_dataset, ind_rad


def process_fft(procstatus, dscfg, radar_list=None):
    """
    Compute the Doppler spectra form the IQ data with a Fourier transform

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted configuration keywords::

        datatype : list of string. Dataset keyword
            The input data types, must contain,
            "IQNdBADUh" and/or "IQNdBADUv" and/or
            "IQNADUh" and/or "IQNADUv" (see new_dataset below)
        window : list of str
            Parameters of the window used to obtain the spectra. The
            parameters are the ones corresponding to function
            scipy.signal.windows.get_window. It can also be ['None'].
    radar_list : list of spectra objects
        Optional. list of spectra objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output fields
        "ShhADUu" (unfiltered_complex_spectra_hh_ADU) if IQNdBADUh was provided,
        "SvvADUu" (unfiltered_complex_spectra_vv_ADU) if IQNdBADUv was provided,
        "sNADUh" (spectral_noise_power_hh_ADU) if IQNADUh was provided,
        "sNADUv" (spectral_noise_power_vv_ADU) if IQNADUv was provided,
    ind_rad : int
        radar index

    """

    if procstatus != 1:
        return None, None

    radarnr, _, datatype, _, _ = get_datatype_fields(dscfg["datatype"][0])
    ind_rad = int(radarnr[5:8]) - 1
    if (radar_list is None) or (radar_list[ind_rad] is None):
        warn("ERROR: No valid radar")
        return None, None
    radar = radar_list[ind_rad]

    wind_params = dscfg.get("window", ["None"])
    if len(wind_params) == 1:
        window = wind_params[0]
        if window == "None":
            window = None
        else:
            try:
                window = float(window)
            except ValueError:
                pass
    else:
        window = wind_params
        for i in range(1, len(window)):
            window[i] = float(window[i])
        window = tuple(window)

    fields_in_list = []
    fields_out_list = []
    for datatypedescr in dscfg["datatype"]:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        field_name = get_fieldname_pyart(datatype)
        if field_name not in radar.fields:
            warn(field_name + " not in radar")
            continue
        if field_name == "IQ_hh_ADU":
            fields_out_list.append("unfiltered_complex_spectra_hh_ADU")
        elif field_name == "IQ_vv_ADU":
            fields_out_list.append("unfiltered_complex_spectra_vv_ADU")
        elif field_name == "IQ_noise_power_hh_ADU":
            fields_out_list.append("spectral_noise_power_hh_ADU")
        elif field_name == "IQ_noiseADU_vv":
            fields_out_list.append("spectral_noise_power_vv_ADU")
        else:
            warn(field_name + " can not be Fourier transformed")
        fields_in_list.append(field_name)

    radar_out = pyart.retrieve.compute_spectra(
        radar, fields_in_list, fields_out_list, window=window
    )

    # prepare for exit
    new_dataset = {"radar_out": radar_out}

    return new_dataset, ind_rad



"""
pyrad.proc.process_icon
===========================

Functions to manage icon data

.. autosummary::
    :toctree: generated/

    process_icon
    process_hzt
    process_iso0_mf
    process_iso0_grib
    process_icon_lookup_table
    process_hzt_lookup_table
    process_icon_to_radar
    process_icon_coord
    process_hzt_coord

"""

from copy import deepcopy
from warnings import warn
import glob
import os

import numpy as np
from netCDF4 import num2date

import pyart

from ..io.io_aux import get_datatype_fields, find_raw_icon_file
from ..io.io_aux import find_hzt_file, find_iso0_file, find_iso0_grib_file
from ..io.io_aux import get_fieldname_pyart
from ..io.read_data_icon import read_icon_data, read_icon_coord
from ..io.read_data_icon import icon2radar_data, icon2radar_coord
from ..io.read_data_icon import get_icon_fields
from ..io.read_data_radar import interpol_field
from ..io.read_data_hzt import read_hzt_data, hzt2radar_data, hzt2radar_coord
from ..io.read_data_hzt import get_iso0_field
from ..io.read_data_iso0_mf import read_iso0_mf_data, read_iso0_grib_data
from ..io.read_data_iso0_mf import iso2radar_data, grib2radar_data

# from memory_profiler import profile


def process_icon(procstatus, dscfg, radar_list=None):
    """
    Gets icon data and put it in radar coordinates

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted Configuration Keywords::

        datatype : string. Dataset keyword
            arbitrary data type supported by pyrad
        keep_in_memory : int. Dataset keyword
            if set keeps the icon data dict, the icon coordinates dict and
            the icon field in radar coordinates in memory
        regular_grid : int. Dataset keyword
            if set it is assume that the radar has a grid constant in time and
            there is no need to compute a new icon field if the icon
            data has not changed
        icon_type : str. Dataset keyword
            name of the icon field to process. Default TEMP
        icon_variables : list of strings. Dataset keyword
            Py-art name of the icon fields. Default temperature
    radar_list : list of Radar objects
        Optional. list of radar objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output field corresponding to
        icon_variables
    ind_rad : int
        radar index

    """
    if procstatus != 1:
        return None, None

    # debugging
    # start_time = time.time()

    for datatypedescr in dscfg["datatype"]:
        radarnr, _, _, _, _ = get_datatype_fields(datatypedescr)
        break

    ind_rad = int(radarnr[5:8]) - 1
    if radar_list[ind_rad] is None:
        warn("No valid radar")
        return None, None
    radar = radar_list[ind_rad]

    keep_in_memory = dscfg.get("keep_in_memory", 0)
    regular_grid = dscfg.get("regular_grid", 0)

    icon_type = dscfg.get("icon_type", "TEMP")
    if icon_type == "TEMP":
        field_names = ["temperature"]
        if "icon_variables" in dscfg:
            field_names = []
            for var in dscfg["icon_variables"]:
                field_names.append(get_fieldname_pyart(var))
        zmin = None
    elif icon_type == "WIND":
        field_names = ["wind_speed", "wind_direction", "vertical_wind_shear"]
        if "icon_variables" in dscfg:
            field_names = []
            for var in dscfg["icon_variables"]:
                field_names.append(get_fieldname_pyart(var))
        zmin = None
    else:
        warn("Unknown icon data type " + icon_type)
        return None, None

    fname = find_raw_icon_file(dscfg["timeinfo"], icon_type, dscfg, ind_rad=ind_rad)

    if fname is None:
        return None, None

    model = os.path.basename(fname)[14:26]
    if model not in ("icon-ch1-eps", "icon-ch2-eps"):
        warn("Unknown NWP model " + model)
        return None, None

    if keep_in_memory:
        if dscfg["initialized"] == 0:
            icon_coord = read_icon_coord(
                dscfg["iconpath"][ind_rad] + "rad2icon/" + model + "_MDR_3D_const.nc",
                zmin=zmin,
            )
            dscfg["global_data"] = {
                "icon_fname": None,
                "icon_data": None,
                "icon_fields": None,
                "time_index": None,
                "icon_coord": icon_coord,
            }
            dscfg["initialized"] = 1

        icon_coord = dscfg["global_data"]["icon_coord"]
        if fname != dscfg["global_data"]["icon_fname"]:
            # debugging
            # start_time2 = time.time()
            icon_data = read_icon_data(fname, field_names=field_names, celsius=True)
            # print(" reading icon takes %s seconds " %
            #      (time.time() - start_time2))
            if icon_data is None:
                warn("icon data not found")
                return None, None

            dscfg["global_data"]["icon_data"] = icon_data
            dscfg["global_data"]["icon_fname"] = fname
        else:
            print("raw icon data already in memory")
            icon_data = dscfg["global_data"]["icon_data"]
    else:
        icon_coord = read_icon_coord(
            dscfg["iconpath"][ind_rad] + "rad2icon/" + model + "_MDR_3D_const.nc",
            zmin=zmin,
        )

        # debugging
        # start_time2 = time.time()
        icon_data = read_icon_data(fname, field_names=field_names, celsius=True)
        # print(" reading icon takes %s seconds " %
        #      (time.time() - start_time2))
        if icon_data is None:
            warn("icon data not found")
            return None, None

    dticon = num2date(icon_data["time"]["data"][:], icon_data["time"]["units"])
    time_index = np.argmin(abs(dticon - dscfg["timeinfo"]))

    if keep_in_memory and regular_grid:
        if time_index != dscfg["global_data"]["time_index"]:
            icon_fields = icon2radar_data(
                radar,
                icon_coord,
                icon_data,
                time_index=time_index,
                field_names=field_names,
            )
            if icon_fields is None:
                warn("Unable to obtain icon fields")
                return None, None

            dscfg["global_data"]["time_index"] = time_index
            dscfg["global_data"]["icon_fields"] = icon_fields
        else:
            print("icon field already in memory")
            icon_fields = dscfg["global_data"]["icon_fields"]
    else:
        icon_fields = icon2radar_data(
            radar, icon_coord, icon_data, time_index=time_index, field_names=field_names
        )
        if icon_fields is None:
            warn("Unable to obtain icon fields")
            return None, None

    # prepare for exit
    new_dataset = {"radar_out": deepcopy(radar)}
    new_dataset["radar_out"].fields = dict()

    for field in icon_fields:
        for field_name in field:
            new_dataset["radar_out"].add_field(field_name, field[field_name])

    return new_dataset, ind_rad


def process_hzt(procstatus, dscfg, radar_list=None):
    """
    Gets iso0 degree data in HZT format and put it in radar coordinates

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted Configuration Keywords::

        metranet_read_lib : str. Global keyword
            Type of METRANET reader library used to read the data.
            Can be 'C' or 'python'
        datatype : string. Dataset keyword
            arbitrary data type supported by pyrad
        keep_in_memory : int. Dataset keyword
            if set keeps the icon data dict, the icon coordinates dict and
            the icon field in radar coordinates in memory
        regular_grid : int. Dataset keyword
            if set it is assume that the radar has a grid constant in time and
            there is no need to compute a new icon field if the icon
            data has not changed
        icon_type : str. Dataset keyword
            name of the icon field to process. Default TEMP
        icon_variables : list of strings. Dataset keyword
            Py-art name of the icon fields. Default temperature
    radar_list : list of Radar objects
        Optional. list of radar objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output fields corresponding to
        icon_variables
    ind_rad : int
        radar index

    """
    if procstatus != 1:
        return None, None

    # debugging
    # start_time = time.time()

    for datatypedescr in dscfg["datatype"]:
        radarnr, _, _, _, _ = get_datatype_fields(datatypedescr)
        break

    ind_rad = int(radarnr[5:8]) - 1
    if radar_list[ind_rad] is None:
        warn("No valid radar")
        return None, None
    radar = radar_list[ind_rad]

    keep_in_memory = dscfg.get("keep_in_memory", 0)
    regular_grid = dscfg.get("regular_grid", 0)

    fname = find_hzt_file(dscfg["timeinfo"], dscfg, ind_rad=ind_rad)

    if fname is None:
        return None, None

    if keep_in_memory:
        if dscfg["initialized"] == 0:
            hzt_data = read_hzt_data(fname, read_lib=dscfg["metranet_read_lib"])
            if hzt_data is None:
                warn("HZT data not found")
                return None, None
            hzt_coord = {"x": hzt_data["x"], "y": hzt_data["y"]}
            dscfg["global_data"] = {
                "hzt_fname": None,
                "hzt_data": None,
                "iso0_field": None,
                "time_index": None,
                "hzt_coord": hzt_coord,
            }
            dscfg["initialized"] = 1

        hzt_coord = dscfg["global_data"]["hzt_coord"]
        if fname != dscfg["global_data"]["hzt_fname"]:
            hzt_data = read_hzt_data(fname, read_lib=dscfg["metranet_read_lib"])
            if hzt_data is None:
                warn("HZT data not found")
                return None, None

            dscfg["global_data"]["hzt_data"] = hzt_data
            dscfg["global_data"]["hzt_fname"] = fname
        else:
            print("raw HZT data already in memory")
            hzt_data = dscfg["global_data"]["hzt_data"]
    else:
        hzt_data = read_hzt_data(fname, read_lib=dscfg["metranet_read_lib"])
        if hzt_data is None:
            warn("HZT data not found")
            return None, None
        hzt_coord = {"x": hzt_data["x"], "y": hzt_data["y"]}

    dthzt = num2date(hzt_data["time"]["data"][:], hzt_data["time"]["units"])
    time_index = np.argmin(abs(dthzt - dscfg["timeinfo"]))

    if keep_in_memory and regular_grid:
        if time_index != dscfg["global_data"]["time_index"]:
            iso0_field = hzt2radar_data(radar, hzt_coord, hzt_data)
            if iso0_field is None:
                warn("Unable to obtain heigth over iso0 field")
                return None, None

            dscfg["global_data"]["time_index"] = time_index
            dscfg["global_data"]["iso0_field"] = iso0_field
        else:
            print("HZT field already in memory")
            iso0_field = dscfg["global_data"]["iso0_field"]
    else:
        iso0_field = hzt2radar_data(radar, hzt_coord, hzt_data)
        if iso0_field is None:
            warn("Unable to obtain HZT fields")
            return None, None

    # prepare for exit
    new_dataset = {"radar_out": deepcopy(radar)}
    new_dataset["radar_out"].fields = dict()
    new_dataset["radar_out"].add_field("height_over_iso0", iso0_field)

    return new_dataset, ind_rad


def process_iso0_mf(procstatus, dscfg, radar_list=None):
    """
    Gets iso0 degree data in text format and put it in radar coordinates.
    This function is meant to process data received from the MeteoFrance NWP
    model. The model provides a maximum of 9 points at 0.5 degree lat/lon
    spacing surrounding a given radar. If a point is not provided it means
    that the iso0 for that point is at or below the ground level. Out of these
    points a single reference iso-0 is obtained according to the user defined
    iso0 statistic.

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted Configuration Keywords::

        datatype : string. Dataset keyword
            arbitrary data type supported by prad
        iso0_statistic : str. Dataset keyword
            The statistic used to weight the iso0 points. Can be avg_by_dist,
            avg, min, max

    radar_list : list of Radar objects
        Optional. list of radar objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output field "H_ISO0"
    ind_rad : int
        radar index

    """
    if procstatus != 1:
        return None, None

    # debugging
    # start_time = time.time()

    for datatypedescr in dscfg["datatype"]:
        radarnr, _, _, _, _ = get_datatype_fields(datatypedescr)
        break

    ind_rad = int(radarnr[5:8]) - 1
    if radar_list[ind_rad] is None:
        warn("No valid radar")
        return None, None
    radar = radar_list[ind_rad]

    iso0_statistic = dscfg.get("iso0_statistic", "avg_by_dist")
    if iso0_statistic not in ("avg_by_dist", "avg", "min", "max"):
        warn(
            "iso0 statistic "
            + iso0_statistic
            + " unknown. Default avg_by_dist will be applied"
        )
        iso0_statistic = "avg_by_dist"

    fname = find_iso0_file(dscfg["timeinfo"], dscfg, ind_rad=ind_rad)
    if fname is None:
        return None, None

    iso0_data = read_iso0_mf_data(fname)
    if iso0_data is None:
        warn("iso0 data not found")
        return None, None

    np.argmin(abs(iso0_data["fcst_time"] - dscfg["timeinfo"]))
    iso0_field = iso2radar_data(
        radar, iso0_data, dscfg["timeinfo"], iso0_statistic=iso0_statistic
    )
    if iso0_field is None:
        warn("Unable to obtain iso0 fields")
        return None, None

    # prepare for exit
    new_dataset = {"radar_out": deepcopy(radar)}
    new_dataset["radar_out"].fields = dict()
    new_dataset["radar_out"].add_field("height_over_iso0", iso0_field)

    return new_dataset, ind_rad


def process_iso0_grib(procstatus, dscfg, radar_list=None):
    """
    Gets iso0 degree data in GRIB format and put it in radar coordinates.
    This function is meant to process data received from the MeteoFrance NWP
    model. It can output the height over the iso0 of each gate or the iso0
    height at each gate

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted Configuration Keywords::

        datatype : string. Dataset keyword
            arbitrary data type supported by pyrad
        time_interp : bool. Dataset keyword
            whether to perform an interpolation in time between consecutive
            model outputs. Default True
        voltype: str. Dataset keyword
            The type of data to output. Can be H_ISO0 or HZT. Default H_ISO0

    radar_list : list of Radar objects
        Optional. list of radar objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output field H_ISO0
    ind_rad : int
        radar index

    """
    if procstatus != 1:
        return None, None

    # debugging
    # start_time = time.time()

    for datatypedescr in dscfg["datatype"]:
        radarnr, _, _, _, _ = get_datatype_fields(datatypedescr)
        break

    ind_rad = int(radarnr[5:8]) - 1
    if radar_list[ind_rad] is None:
        warn("No valid radar")
        return None, None
    radar = radar_list[ind_rad]

    time_interp = dscfg.get("time_interp", True)
    voltype = dscfg.get("voltype", "H_ISO0")
    field_name = get_fieldname_pyart(voltype)

    fname = find_iso0_grib_file(dscfg["timeinfo"], dscfg, ind_rad=ind_rad)
    if fname is None:
        return None, None

    iso0_data = read_iso0_grib_data(fname)
    if iso0_data is None:
        warn("iso0 data not found")
        return None, None

    np.argmin(abs(iso0_data["fcst_time"] - dscfg["timeinfo"]))
    iso0_field = grib2radar_data(
        radar,
        iso0_data,
        dscfg["timeinfo"],
        time_interp=time_interp,
        field_name=field_name,
    )
    if iso0_field is None:
        warn("Unable to obtain iso0 fields")
        return None, None

    # prepare for exit
    new_dataset = {"radar_out": deepcopy(radar)}
    new_dataset["radar_out"].fields = dict()
    new_dataset["radar_out"].add_field(field_name, iso0_field)

    return new_dataset, ind_rad


# @profile
def process_icon_lookup_table(procstatus, dscfg, radar_list=None):
    """
    Gets icon data and put it in radar coordinates
    using look up tables computed or loaded when initializing

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted Configuration Keywords::

        datatype : string. Dataset keyword
            arbitrary data type supported by pyrad
        lookup_table : int. Dataset keyword
            if set a pre-computed look up table for the icon coordinates is
            loaded. Otherwise the look up table is computed taking the first
            radar object as reference
        regular_grid : int. Dataset keyword
            if set it is assume that the radar has a grid constant in time and
            therefore there is no need to interpolate the icon field in
            memory to the current radar grid
        icon_type : str. Dataset keyword
            name of the icon field to process. Default TEMP
        icon_variables : list of strings. Dataset keyword
            Py-art name of the icon fields. Default temperature
    radar_list : list of Radar objects
        Optional. list of radar objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output fields corresponding to icon_variables
    ind_rad : int
        radar index

    """
    if procstatus != 1:
        return None, None

    # debugging
    # start_time = time.time()

    for datatypedescr in dscfg["datatype"]:
        radarnr, _, _, _, _ = get_datatype_fields(datatypedescr)
        break

    ind_rad = int(radarnr[5:8]) - 1
    if radar_list[ind_rad] is None:
        warn("No valid radar")
        return None, None
    radar = radar_list[ind_rad]

    regular_grid = dscfg.get("regular_grid", 0)
    lookup_table = dscfg.get("lookup_table", 0)

    icon_type = dscfg.get("icon_type", "TEMP")
    if icon_type == "TEMP":
        field_names = ["temperature"]
        if "icon_variables" in dscfg:
            field_names = []
            for var in dscfg["icon_variables"]:
                field_names.append(get_fieldname_pyart(var))
        zmin = None
    elif icon_type == "WIND":
        field_names = ["wind_speed", "wind_direction", "vertical_wind_shear"]
        if "icon_variables" in dscfg:
            field_names = []
            for var in dscfg["icon_variables"]:
                field_names.append(get_fieldname_pyart(var))
        zmin = None
    else:
        warn("Unknown icon data type " + icon_type)
        return None, None

    fname = find_raw_icon_file(dscfg["timeinfo"], icon_type, dscfg, ind_rad=ind_rad)

    if fname is None:
        return None, None

    model = os.path.basename(fname)[14:26]
    if model not in ("icon-ch1-eps", "icon-ch2-eps"):
        warn("Unknown NWP model " + model)
        return None, None

    if dscfg["initialized"] == 0:
        if lookup_table:
            savedir = dscfg["iconpath"][ind_rad] + "rad2icon/"
            fname_ind = "rad2icon_icon_index_" + dscfg["procname"] + ".nc"
            fname_ind2 = glob.glob(savedir + fname_ind)
            if not fname_ind2:
                warn("File " + savedir + fname_ind + " not found")
                return None, None
            icon_radar = pyart.io.read_cfradial(fname_ind2[0])
        else:
            icon_coord = read_icon_coord(
                dscfg["iconpath"][ind_rad] + "rad2icon/" + model + "_MDR_3D_const.nc",
                zmin=zmin,
            )
            print("icon coordinates files read")
            icon_ind_field = icon2radar_coord(radar, icon_coord)
            icon_radar = deepcopy(radar)
            icon_radar.fields = dict()
            icon_radar.add_field("icon_index", icon_ind_field)
            print("icon index field added")

        dscfg["global_data"] = {
            "icon_fname": None,
            "icon_data": None,
            "icon_fields": None,
            "icon_radar": icon_radar,
            "time_index": None,
        }
        dscfg["initialized"] = 1

    if fname != dscfg["global_data"]["icon_fname"]:
        # debugging
        # start_time2 = time.time()
        icon_data = read_icon_data(fname, field_names=field_names, celsius=True)
        # print(" reading icon takes %s seconds " %
        #      (time.time() - start_time2))
        if icon_data is None:
            warn("icon data not found")
            return None, None

        dscfg["global_data"]["icon_data"] = icon_data
    else:
        print("raw icon data already in memory")
        icon_data = dscfg["global_data"]["icon_data"]

    dticon = num2date(icon_data["time"]["data"][:], icon_data["time"]["units"])

    time_index = np.argmin(abs(dticon - dscfg["timeinfo"]))

    if (
        fname != dscfg["global_data"]["icon_fname"]
        or time_index != dscfg["global_data"]["time_index"]
    ):
        # debugging
        # start_time3 = time.time()
        icon_fields = get_icon_fields(
            icon_data,
            dscfg["global_data"]["icon_radar"].fields["icon_index"],
            time_index=time_index,
            field_names=field_names,
        )
        if icon_fields is None:
            warn("Unable to obtain icon fields")
            return None, None
        # print(" getting icon data takes %s seconds "
        #      % (time.time() - start_time3))

        dscfg["global_data"]["time_index"] = time_index
        dscfg["global_data"]["icon_fields"] = icon_fields
    else:
        print("icon field already in memory")
        icon_fields = dscfg["global_data"]["icon_fields"]

    dscfg["global_data"]["icon_fname"] = fname

    # prepare for exit
    new_dataset = {"radar_out": deepcopy(radar)}
    new_dataset["radar_out"].fields = dict()

    if not regular_grid:
        radar_aux = deepcopy(dscfg["global_data"]["icon_radar"])
        radar_aux.fields = dict()

    for field in icon_fields:
        for field_name in field:
            try:
                if regular_grid:
                    new_dataset["radar_out"].add_field(field_name, field[field_name])
                else:
                    # interpolate to current radar grid
                    radar_aux.add_field(field_name, field[field_name])
                    icon_field_interp = interpol_field(radar, radar_aux, field_name)
                    new_dataset["radar_out"].add_field(field_name, icon_field_interp)
            except Exception as ee:
                warn(str(ee))
                warn("Unable to add icon " + field_name + " field to radar object")
                return None, None

    return new_dataset, ind_rad


def process_hzt_lookup_table(procstatus, dscfg, radar_list=None):
    """
    Gets HZT data and put it in radar coordinates
    using look up tables computed or loaded when initializing

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted Configuration Keywords::

        metranet_read_lib : str. Global keyword
            Type of METRANET reader library used to read the data.
            Can be 'C' or 'python'
        datatype : string. Dataset keyword
            arbitrary data type supported by pyrad
        lookup_table : int. Dataset keyword
            if set a pre-computed look up table for the icon coordinates is
            loaded. Otherwise the look up table is computed taking the first
            radar object as reference
        regular_grid : int. Dataset keyword
            if set it is assume that the radar has a grid constant in time and
            therefore there is no need to interpolate the icon field in
            memory to the current radar grid
    radar_list : list of Radar objects
        Optional. list of radar objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output field HISO0
    ind_rad : int
        radar index

    """
    if procstatus != 1:
        return None, None

    # debugging
    # start_time = time.time()

    for datatypedescr in dscfg["datatype"]:
        radarnr, _, _, _, _ = get_datatype_fields(datatypedescr)
        break

    ind_rad = int(radarnr[5:8]) - 1
    if radar_list[ind_rad] is None:
        warn("No valid radar")
        return None, None
    radar = radar_list[ind_rad]

    regular_grid = dscfg.get("regular_grid", 0)
    lookup_table = dscfg.get("lookup_table", 0)

    fname = find_hzt_file(dscfg["timeinfo"], dscfg, ind_rad=ind_rad)

    if fname is None:
        return None, None

    if dscfg["initialized"] == 0:
        if lookup_table:
            savedir = dscfg["iconpath"][ind_rad] + "rad2icon/"
            fname_ind = "rad2icon_hzt_index_" + dscfg["procname"] + ".nc"
            fname_ind2 = glob.glob(savedir + fname_ind)
            if not fname_ind2:
                warn("File " + savedir + fname_ind + " not found")
                return None, None
            hzt_radar = pyart.io.read_cfradial(fname_ind2[0])
        else:
            hzt_data = read_hzt_data(fname, read_lib=dscfg["metranet_read_lib"])
            if hzt_data is None:
                warn("HZT data not found")
                return None, None
            hzt_coord = {"x": hzt_data["x"], "y": hzt_data["y"]}
            hzt_ind_field = hzt2radar_coord(radar, hzt_coord)
            hzt_radar = deepcopy(radar)
            hzt_radar.fields = dict()
            hzt_radar.add_field("hzt_index", hzt_ind_field)

        dscfg["global_data"] = {
            "hzt_fname": None,
            "hzt_data": None,
            "iso0_field": None,
            "hzt_radar": hzt_radar,
            "time_index": None,
        }
        dscfg["initialized"] = 1

    if fname != dscfg["global_data"]["hzt_fname"]:
        hzt_data = read_hzt_data(fname, read_lib=dscfg["metranet_read_lib"])
        if hzt_data is None:
            warn("HZT data not found")
            return None, None

        dscfg["global_data"]["hzt_data"] = hzt_data
    else:
        print("HZT data already in memory")
        hzt_data = dscfg["global_data"]["hzt_data"]

    dthzt = num2date(hzt_data["time"]["data"][:], hzt_data["time"]["units"])
    time_index = np.argmin(abs(dthzt - dscfg["timeinfo"]))

    if (
        fname != dscfg["global_data"]["hzt_fname"]
        or time_index != dscfg["global_data"]["time_index"]
    ):
        # debugging
        # start_time3 = time.time()
        iso0_field = get_iso0_field(
            hzt_data,
            dscfg["global_data"]["hzt_radar"].fields["hzt_index"],
            dscfg["global_data"]["hzt_radar"].gate_altitude["data"],
        )
        if iso0_field is None:
            warn("Unable to obtain height over iso0")
            return None, None

        dscfg["global_data"]["time_index"] = time_index
        dscfg["global_data"]["iso0_field"] = iso0_field
    else:
        print("HZT field already in memory")
        iso0_field = dscfg["global_data"]["iso0_field"]

    dscfg["global_data"]["hzt_fname"] = fname

    # prepare for exit
    new_dataset = {"radar_out": deepcopy(radar)}
    new_dataset["radar_out"].fields = dict()

    if not regular_grid:
        radar_aux = deepcopy(dscfg["global_data"]["hzt_radar"])
        radar_aux.fields = dict()

    try:
        if regular_grid:
            new_dataset["radar_out"].add_field("height_over_iso0", iso0_field)
        else:
            # interpolate to current radar grid
            radar_aux.add_field("height_over_iso0", iso0_field)
            hzt_field_interp = interpol_field(radar, radar_aux, "height_over_iso0")
            new_dataset["radar_out"].add_field("height_over_iso0", hzt_field_interp)
    except Exception as ee:
        warn(str(ee))
        warn("Unable to add height_over_iso0 " + " field to radar object")
        return None, None

    return new_dataset, ind_rad


# @profile
def process_icon_to_radar(procstatus, dscfg, radar_list=None):
    """
    Gets icon data and put it in radar coordinates using look up tables

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted Configuration Keywords::

        datatype : string. Dataset keyword
            arbitrary data type supported by pyrad
        icon_type : str. Dataset keyword
            name of the icon field to process. Default TEMP
        icon_variables : list of strings. Dataset keyword
            Py-art name of the icon fields. Default temperature
        icon_time_index_min, icon_time_index_max : int
            minimum and maximum indices of the icon data to retrieve. If a
            value is provided only data corresponding to the time indices
            within the interval will be used. If None all data will be used.
            Default None
    radar_list : list of Radar objects
        Optional. list of radar objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output fields corresponding to icon_variables
    ind_rad : int
        radar index

    """
    if procstatus != 1:
        return None, None

    for datatypedescr in dscfg["datatype"]:
        radarnr, _, _, _, _ = get_datatype_fields(datatypedescr)
        break

    ind_rad = int(radarnr[5:8]) - 1

    icon_type = dscfg.get("icon_type", "TEMP")
    if icon_type == "TEMP":
        field_names = ["temperature"]
        if "icon_variables" in dscfg:
            field_names = []
            for var in dscfg["icon_variables"]:
                field_names.append(get_fieldname_pyart(var))
    elif icon_type == "WIND":
        field_names = ["wind_speed", "wind_direction", "vertical_wind_shear"]
        if "icon_variables" in dscfg:
            field_names = []
            for var in dscfg["icon_variables"]:
                field_names.append(get_fieldname_pyart(var))
    else:
        warn("Unknown icon data type " + icon_type)
        return None, None

    time_index_min = dscfg.get("icon_time_index_min", None)
    time_index_max = dscfg.get("icon_time_index_max", None)

    fname = find_raw_icon_file(dscfg["timeinfo"], icon_type, dscfg, ind_rad=ind_rad)

    if fname is None:
        return None, None

    model = os.path.basename(fname)[14:26]
    if model not in ("icon-ch1-eps", "icon-ch2-eps"):
        warn("Unknown NWP model " + model)
        return None, None

    if dscfg["initialized"] == 0:
        savedir = dscfg["iconpath"][ind_rad] + "rad2icon/"
        fname_ind = "rad2icon_icon_index_" + dscfg["procname"] + ".nc"
        fname_ind2 = glob.glob(savedir + fname_ind)
        if not fname_ind2:
            warn("File " + savedir + fname_ind + " not found")
            return None, None
        icon_radar = pyart.io.read_cfradial(fname_ind2[0])

        dscfg["global_data"] = {"icon_radar": icon_radar}
        dscfg["initialized"] = 1

    icon_data = read_icon_data(fname, field_names=field_names, celsius=True)
    if icon_data is None:
        warn("icon data not found")
        return None, None

    dticon = num2date(icon_data["time"]["data"][:], icon_data["time"]["units"])

    if time_index_min is None:
        time_index_min = 0
    if time_index_max is None:
        time_index_max = len(dticon) - 1

    if time_index_max > len(dticon) - 1:
        warn(
            "icon_time_index_max larger than number of icon forecasts"
            "icon forecasts " + len(dticon)
        )
        time_index_max = len(dticon) - 1

    icon_radars = []
    for time_index, dtc in enumerate(dticon):
        if time_index < time_index_min:
            continue
        if time_index > time_index_max:
            break

        icon_fields = get_icon_fields(
            icon_data,
            dscfg["global_data"]["icon_radar"].fields["icon_index"],
            time_index=time_index,
            field_names=field_names,
        )
        if icon_fields is None:
            warn("Unable to obtain icon fields")
            return None, None

        radar_out = deepcopy(dscfg["global_data"]["icon_radar"])
        radar_out.fields = dict()
        radar_out.time["units"] = dtc.strftime("seconds since %Y-%m-%dT%H:%M:%SZ")

        for field in icon_fields:
            for field_name in field:
                radar_out.add_field(field_name, field[field_name])

        icon_radars.append({"ind_rad": ind_rad, "radar_out": radar_out, "dticon": dtc})

    return icon_radars, ind_rad


def process_icon_coord(procstatus, dscfg, radar_list=None):
    """
    Gets the icon indices corresponding to each icon coordinates

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted Configuration Keywords::

        datatype : string. Dataset keyword
            arbitrary data type supported by pyrad
        iconpath : string. General keyword
            path where to store the look up table
        model : string. Dataset keyword
            The icon model to use. Can be icon-1, icon-1e, icon-2, icon-7
    radar_list : list of Radar objects
        Optional. list of radar objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output field
        "icon_index"
    ind_rad : int
        radar index

    """
    if procstatus != 1:
        return None, None

    if dscfg["initialized"] == 1:
        return None, None

    for datatypedescr in dscfg["datatype"]:
        radarnr, _, _, _, _ = get_datatype_fields(datatypedescr)
        break

    ind_rad = int(radarnr[5:8]) - 1
    if radar_list[ind_rad] is None:
        warn("No valid radar")
        return None, None
    radar = radar_list[ind_rad]

    model = dscfg.get("model", "icon-ch1-eps")
    if model not in ("icon-ch1-eps", "icon-ch2-eps"):
        warn("Unknown NWP model " + model)
        return None, None

    icon_coord = read_icon_coord(
        dscfg["iconpath"][ind_rad] + "rad2icon/" + model + "_MDR_3D_const.nc", zmin=None
    )

    if icon_coord is None:
        return None, None

    icon_ind_field = icon2radar_coord(radar, icon_coord, slice_xy=True, slice_z=False)

    # prepare for exit
    radar_obj = deepcopy(radar)
    radar_obj.fields = dict()
    radar_obj.add_field("icon_index", icon_ind_field)

    new_dataset = {"ind_rad": ind_rad, "radar_out": radar_obj}

    dscfg["initialized"] = 1

    return new_dataset, ind_rad


def process_hzt_coord(procstatus, dscfg, radar_list=None):
    """
    Gets the HZT indices corresponding to each HZT coordinates

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted Configuration Keywords::

        metranet_read_lib : str. Global keyword
            Type of METRANET reader library used to read the data.
            Can be 'C' or 'python'
        datatype : string. Dataset keyword
            arbitrary data type supported by pyrad
        iconpath : string. General keyword
            path where to store the look up table
    radar_list : list of Radar objects
        Optional. list of radar objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output field "icon_index"
    ind_rad : int
        radar index

    """
    if procstatus != 1:
        return None, None

    if dscfg["initialized"] == 1:
        return None, None

    for datatypedescr in dscfg["datatype"]:
        radarnr, _, _, _, _ = get_datatype_fields(datatypedescr)
        break

    ind_rad = int(radarnr[5:8]) - 1
    if radar_list[ind_rad] is None:
        warn("No valid radar")
        return None, None
    radar = radar_list[ind_rad]

    fname = find_hzt_file(dscfg["timeinfo"], dscfg, ind_rad=ind_rad)

    hzt_data = read_hzt_data(fname, read_lib=dscfg["metranet_read_lib"])
    if hzt_data is None:
        warn("HZT data not found")
        return None, None
    hzt_coord = {"x": hzt_data["x"], "y": hzt_data["y"]}

    hzt_ind_field = hzt2radar_coord(radar, hzt_coord)

    # prepare for exit
    radar_obj = deepcopy(radar)
    radar_obj.fields = dict()
    radar_obj.add_field("hzt_index", hzt_ind_field)

    new_dataset = {"ind_rad": ind_rad, "radar_out": radar_obj}

    dscfg["initialized"] = 1

    return new_dataset, ind_rad



"""
pyrad.prod.process_product
==========================

Functions for obtaining Pyrad products from the datasets

.. autosummary::
    :toctree: generated/

    generate_traj_product

"""

from warnings import warn

from ..io.io_aux import get_save_dir, make_filename
from ..io.timeseries import TimeSeries
from ..graph.plots import plot_pos


def generate_traj_product(traj, prdcfg):
    """
    Generates trajectory products. Accepted product types:
        'TRAJ_MAP': Plots the trajectory on a lat-lon map with the altitude
            color coded
        'TRAJ_PLOT': Plots time series of the trajectory respect to the radar
            elevation, azimuth or range
            User defined parameters:
                'datatype': str
                    The type of parameter: 'EL', 'AZ', or 'RANGE'
        'TRAJ_TEXT': Writes the trajectory information in a csv file

    Parameters
    ----------
    traj : Trajectory object

    prdcfg : dictionary of dictionaries
        product configuration dictionary of dictionaries

    Returns
    -------
    None

    """

    dssavedir = prdcfg["dsname"]
    if "dssavename" in prdcfg:
        dssavedir = prdcfg["dssavename"]

    if prdcfg["type"] == "TRAJ_PLOT":

        timeinfo = traj.time_vector[0]

        savedir = get_save_dir(
            prdcfg["basepath"],
            prdcfg["procname"],
            dssavedir,
            prdcfg["prdname"],
            timeinfo=timeinfo,
        )

        ts = TimeSeries("", traj.time_vector, timeformat="%Y-%m-%d %H:%M:%S.%f")

        if prdcfg["datatype"] == "EL":
            fname = make_filename(
                "ts",
                prdcfg["dstype"],
                "TRAJ",
                prdcfg["imgformat"],
                prdcfginfo="EL",
                timeinfo=timeinfo,
                timeformat="%Y%m%d%H%M%S",
                runinfo=prdcfg["runinfo"],
            )

            ts.add_dataseries(
                "Elevation", "Elevation", "deg", traj.radar_list[0].elevation_vec
            )
            ts.plot(savedir + fname[0])

        elif prdcfg["datatype"] == "AZ":
            fname = make_filename(
                "ts",
                prdcfg["dstype"],
                "TRAJ",
                prdcfg["imgformat"],
                prdcfginfo="AZ",
                timeinfo=timeinfo,
                timeformat="%Y%m%d%H%M%S",
                runinfo=prdcfg["runinfo"],
            )

            ts.add_dataseries(
                "Azimuth", "Azimuth", "deg", traj.radar_list[0].azimuth_vec
            )
            ts.plot(savedir + fname[0])

        elif prdcfg["datatype"] == "RANGE":
            fname = make_filename(
                "ts",
                prdcfg["dstype"],
                "TRAJ",
                prdcfg["imgformat"],
                prdcfginfo="RANGE",
                timeinfo=timeinfo,
                timeformat="%Y%m%d%H%M%S",
                runinfo=prdcfg["runinfo"],
            )

            ts.add_dataseries("Range", "Range", "m", traj.radar_list[0].range_vec)
            ts.plot(savedir + fname[0])

        else:
            raise Exception(
                "ERROR: Unknown datatype '%s' (dataset: '%s')"
                % (prdcfg["datatype"], prdcfg["dsname"])
            )

        return None

    if prdcfg["type"] == "TRAJ_TEXT":

        timeinfo = traj.time_vector[0]

        savedir = get_save_dir(
            prdcfg["basepath"],
            prdcfg["procname"],
            dssavedir,
            prdcfg["prdname"],
            timeinfo=timeinfo,
        )

        fname = make_filename(
            "ts",
            prdcfg["dstype"],
            "TRAJ",
            ["csv"],
            prdcfginfo=None,
            timeinfo=timeinfo,
            timeformat="%Y%m%d%H%M%S",
            runinfo=prdcfg["runinfo"],
        )

        description = [
            "Description:",
            "Time series of a plane trajectory in radar " "coordinates.",
        ]

        ts = TimeSeries(
            description, traj.time_vector, timeformat="%Y-%m-%d %H:%M:%S.%f"
        )
        ts.add_dataseries(
            "Elevation", "Elevation", "deg", traj.radar_list[0].elevation_vec
        )
        ts.add_dataseries("Azimuth", "Azimuth", "deg", traj.radar_list[0].azimuth_vec)
        ts.add_dataseries("Range", "Range", "m", traj.radar_list[0].range_vec)

        ts.add_dataseries(
            "Absolute Speed", "Absolute Speed", "m/s", traj.radar_list[0].v_abs
        )
        ts.add_dataseries("Radial Speed", "Radial Speed", "m/s", traj.radar_list[0].v_r)
        ts.add_dataseries(
            "Elevation Speed", "Elevation Speed", "deg/s", traj.radar_list[0].v_el
        )
        ts.add_dataseries(
            "Azimuth Speed", "Azimuth Speed", "deg/s", traj.radar_list[0].v_az
        )

        ts.write(savedir + fname[0])

        return None

    if prdcfg["type"] == "TRAJ_MAP":  # Trajectory on a map
        timeinfo = traj.time_vector[0]
        savedir = get_save_dir(
            prdcfg["basepath"],
            prdcfg["procname"],
            dssavedir,
            prdcfg["prdname"],
            timeinfo=timeinfo,
        )

        fname = make_filename(
            "ts",
            prdcfg["dstype"],
            "TRAJ",
            prdcfg["imgformat"],
            prdcfginfo="MAP",
            timeinfo=timeinfo,
            timeformat="%Y%m%d%H%M%S",
            runinfo=prdcfg["runinfo"],
        )

        title = "Trajectory Starting at  %s" % traj.time_vector[0].strftime("%Y-%m-%d")

        fname_list = fname
        for i, fname in enumerate(fname_list):
            fname_list[i] = savedir + fname

        # Get traj
        lat = traj.wgs84_lat_deg
        lon = traj.wgs84_lon_deg
        alt = traj.wgs84_alt_m

        plot_pos(
            lat,
            lon,
            alt,
            fname_list,
            cb_label="Altitude [m]",
            titl=title,
            save_fig=True,
        )

        return None

    warn(" Unsupported product type: " + prdcfg["type"])
    return None
